from __future__ import annotations

import re
from click import argument, command, Context, option, Parameter, BadParameter, Option, Argument
from functools import partial, wraps
from re import fullmatch, IGNORECASE
from typing import Sequence, Any, Callable

from utz.rgx import Includes, Excludes, Patterns

INT_RGX = re.compile(r'(?P<base>[\d.]+)(?P<suffix>[kmbg]i?)', flags=IGNORECASE)
ORDERS_SI  = { 'k': 2**10, 'm': 2**20, 'g': 2**30 }
ORDERS_IEC = { 'k': 1e3  , 'm': 1e6  , 'g': 1e9  , 'b': 1e9 }

arg = argument
cmd = command
opt = option


def parse_int(value: str | int | None) -> int:
    if isinstance(value, int):
        return value
    m = fullmatch(INT_RGX, value.lower())
    if m:
        n = float(m['base'])
        suffix = m['suffix']
        if suffix.endswith('i'):
            suffix = suffix[:-1]
            orders = ORDERS_SI
        else:
            orders = ORDERS_IEC
        if suffix not in orders:
            raise ValueError(f'Failed to parse "{value}"')
        order = orders[suffix]
        return int(n * order)
    else:
        try:
            return int(float(value))
        except Exception:
            raise ValueError(f'Failed to parse "{value}"')


def int_cb(ctx: Context, param: Parameter, value: str | int | None) -> int:
    if value is None:
        raise BadParameter(f"expected value, received None", ctx=ctx, param=param)
    try:
        return parse_int(value)
    except Exception:
        raise BadParameter(f'failed to parse "{value}"', ctx=ctx, param=param)


def parse_int_opt(ctx: Context, param: Parameter, value: str) -> int | None:
    if value is None:
        return None
    else:
        return int_cb(ctx, param, value)


def num(
    *args,
    required: bool = False,
    default: int | str | None = None,
    **kwargs,
):
    return option(
        *args,
        required=required,
        default=str(default) if isinstance(default, int) else default,
        callback=int_cb if required or default else parse_int_opt,
        **kwargs,
    )


number = num


def count(
    *args,
    values: Sequence[Any] | None = None,
    **kwargs,
):
    def count_enum_cb(ctx: Context, param: Parameter, value: int):
        if values is not None:
            if value >= len(values):
                raise BadParameter(f"expected [0,{len(values)}), found {value}", ctx=ctx, param=param)
            value = values[value]
        if 'callback' in kwargs:
            return kwargs['callback'](value)
        else:
            return value

    return option(
        *args,
        count=True,
        callback=count_enum_cb,
        **kwargs,
    )


Parse = Callable[[str], Any]


def dict_cb(
    ctx: Context,
    param: Parameter,
    value: tuple[str, ...],
    parse: Parse | None = None,
) -> dict[str, str]:
    """``click.option`` ``callback`` that loads multiple "k=v"-style values into a ``dict``."""
    rv = {}
    for kv in value:
        pcs = kv.split('=', 1)
        if len(pcs) != 2:
            raise BadParameter(f"bad value: {kv}", ctx, param)
        k, v = pcs
        if parse:
            v = parse(v)
        rv[k] = v
    return rv


def obj(
    *args: str,
    parse: Parse | None = None,
    **kwargs,
):
    """``click.option`` wrapper that loads multiple "k=v"-style values into a ``dict``."""
    return option(
        *args,
        multiple=True,
        callback=partial(dict_cb, parse=parse),
        **kwargs,
    )


def multi_cb(
    ctx: Context,
    param: Parameter,
    value: str | tuple[str, ...],
    sep: str | None = ',',
    parse: Parse | None = None,
) -> tuple[str, ...]:
    """``click.option`` ``callback`` that combines multiple ``sep``-delimited strings into a ``tuple``."""
    if isinstance(value, str):
        value = (value,)
    if sep is None:
        return value
    rv = []
    for v in value:
        for s in v.split(sep):
            if parse:
                try:
                    s = parse(s)
                except Exception:
                    raise BadParameter(f'failed to parse "{s}"', ctx=ctx, param=param)
            rv.append(s)
    return tuple(rv)


def multi(
    *args: str,
    sep: str | None = ',',
    parse: Parse | None = None,
    multiple: bool = True,
    **kwargs,
):
    """``click.option`` wrapper that combines multiple ``sep``-delimited strings into a ``tuple``."""
    return opt(
        *args,
        callback=partial(multi_cb, sep=sep, parse=parse),
        multiple=multiple,
        **kwargs,
    )


paths = partial(multi, sep=':')
flag = partial(option, is_flag=True)


def flags(*names, default=None, **kwargs):
    def long_opt(name):
        _default = False
        if name.startswith('no-'):
            name = name[3:]
            _default = True

        arg = f'--{name}/--no-{name}'
        return dict(
            name=name,
            default=_default,
            arg=arg,
        )

    def short_opt(name):
        if len(name) != 1:
            raise ValueError(f'Unrecognized short option name: -{name}')
        _default = name.isupper()
        name = name.lower()
        arg = f'-{name}/-{name.upper()}'
        return dict(
            name=name,
            default=_default,
            arg=arg,
        )

    def make_opt(name):
        if name.startswith('--'):
            return long_opt(name[2:])
        elif name.startswith('-'):
            return short_opt(name[1:])
        elif len(name) > 1:
            return long_opt(name)
        elif len(name) == 1:
            return short_opt(name)
        else:
            raise ValueError('Empty flag')

    opts = [ make_opt(name) for name in names ]
    if default is None:
        defaults = set([ o['default'] for o in opts ])
        if len(defaults) > 1: raise ValueError(f'Conflicting defaults: {opts}')
        [default] = defaults

    return opt(
        *[ o['arg'] for o in opts ],
        default=default,
        **kwargs
    )


Deco = Callable[[Callable], Callable]


def patterns(
    *args: str | Deco,
    cls: type[Patterns],
    flags: int = 0,
    search: bool = True,
    **kwargs,
):
    """Decorator that parses "include" regexs into an ``Includes`` object.

    include_opt:
        ``click.option`` specifying
    kwarg:
        The returned ``Patterns`` object will be passed to the user's final ``click.Command`` function with this
        keyword-arg name.
    flags:
        Regex flags passed to ``Includes`` or ``Excludes``.
    search:
        Passed to ``Includes`` or ``Excludes``, controls use of ``search`` (default) vs. ``fullmatch``.
    """

    if len(args) == 1 and callable(args[0]) and not kwargs:
        deco = args[0]
    else:
        deco = option(*args, **kwargs)

    def rv(fn):

        nonlocal deco
        param: Argument | Option | None = None

        @deco
        @wraps(fn)
        def _fn(*args, **kwargs):
            nonlocal param
            assert param
            name = param.name
            pats = kwargs.pop(name)
            if isinstance(pats, str):
                # Non-`multi` version, e.g. `incs('-i', '--include', 'patterns'),`
                pats = (pats,)
            patterns = cls(pats or None, flags=flags, search=search)
            return fn(
                *args,
                **{ name: patterns },
                **kwargs,
            )

        param = _fn.__click_params__[-1]

        return _fn

    return rv


includes = incs = partial(patterns, cls=Includes)
excludes = excs = partial(patterns, cls=Excludes)


def include_exclude(
    include_opt: Deco,
    exclude_opt: Deco,
    kwarg: str = 'patterns',
    flags: int = 0,
    search: bool = True,
):
    """Decorator that adds two exclusive ``click.option``s, for specifying regexs to include or exclude.

    kwarg:
        The returned ``Patterns`` object will be passed to the user's final ``click.Command`` function with this
        keyword-arg name.
    flags:
        Regex flags passed to ``Includes`` or ``Excludes``.
    search:
        Passed to ``Includes`` or ``Excludes``, controls use of ``search`` (default) vs. ``fullmatch``.
    """
    patterns_kwargs = dict(flags=flags, search=search)

    def rv(fn):

        include_option: Option | None = None
        exclude_option: Option | None = None

        @include_opt
        @exclude_opt
        @wraps(fn)
        def _fn(*args, **kwargs):
            nonlocal include_option, exclude_option
            assert include_option
            assert exclude_option
            includes = kwargs.pop(include_option.name)
            excludes = kwargs.pop(exclude_option.name)
            if includes:
                if excludes:
                    raise ValueError(f"Pass {'/'.join(include_option.opts)} xor {'/'.join(exclude_option.opts)}")
                else:
                    patterns = Includes(includes or None, **patterns_kwargs)
            else:
                patterns = Excludes(excludes or None, **patterns_kwargs)
            return fn(
                *args,
                **{ kwarg: patterns },
                **kwargs,
            )

        include_option = _fn.__click_params__[-1]
        exclude_option = _fn.__click_params__[-2]

        return _fn

    return rv


inc_exc = include_exclude



import json
from pytest import raises
from utz import o


def test_kwargs():
    o1 = o(a=1,b=2)
    assert o1.a == 1
    assert o1.b == 2
    with raises(AttributeError):
        o1.c


def test_dict():
    x={'c':3,'d':4}
    o2 = o(x)

    assert o2.c == 3
    assert o2.d == 4

    # mutations in each direction
    x['e'] = 5
    assert o2.e == 5

    o2.c = 'ccc'
    assert x['c'] == 'ccc'

    # membership
    assert 'c' in o2
    assert not 'z' in o2

    # str/repr
    assert str(o2) == "{'c': 'ccc', 'd': 4, 'e': 5}"
    assert repr(o2) == "{'c': 'ccc', 'd': 4, 'e': 5}"


def test_nested():
    o1 = o(a={'b':1})
    o2 = o(a={'b':1})
    o3 = o(a=o(b=1))

    assert o1 is not o2
    assert o1 is not o3
    assert o2 is not o3

    def eq(l, r):
        assert l == r
        assert r == l

    def ne(l, r):
        assert l != r
        assert r != l

    eq(o1, o2)
    eq(o1, o3)
    eq(o2, o3)

    eq(o1, {'a':{'b':1}})
    eq(o1.a, {'b':1})
    eq(o1.a, o(b=1))

    ne(o1, o(a=o(b=2)))
    ne(o1, o(a=o(b=1,c=3)))
    ne(o1, o(a=o(b=1),c=3))

    assert o1.get('b') is None
    assert o1.get('a') == o(b=1)
    assert o1.get('b', 'default') == 'default'
    assert o1.get('a', 'default') == o(b=1)
    assert o1('a', 'b', default='default') == 1
    assert o1('a', 'c', default='default') == 'default'
    assert o1('a', 'c') == None
    assert o1('a', 'c', 'b') == None

    assert o1.a.b == 1
    with raises(AttributeError):
        o1.a.b.c
    with raises(AttributeError):
        o1.a.c


def test_list():
    _o = o(a=1, b=2)
    assert list(iter(_o)) == [ 'a', 'b' ]
    assert list(_o.items()) == [ ('a',1), ('b',2) ]


def test_serialization():
    _o = o(a=1, b={'c':3})
    _o.update(d=4)
    assert json.dumps(_o) == '{"a": 1, "b": {"c": 3}, "d": 4}'


def test_merge():
    d1 = { 'a':1,'b':2 }
    d2 = { 'a':11,'c':33 }
    o1 = o(d1).merge(d2)
    assert dict(o1) == { 'a':11,'b':2,'c':33 }
    assert (o1.a, o1.b, o1.c) == (11, 2, 33)
    assert d1 == {'a':1,'b':2}
    assert d2 == { 'a':11,'c':33 }


def test_merge_static():
    d1 = { 'a':1,'b':2 }
    d2 = { 'a':11,'c':33 }
    o1 = o.merge(d1, d2)
    assert dict(o1) == { 'a':11,'b':2,'c':33 }
    assert (o1.a, o1.b, o1.c) == (11, 2, 33)
    assert d1 == {'a':1,'b':2}
    assert d2 == { 'a':11,'c':33 }


def test_update():
    o1 = o(a=1,b=2)
    o1.update(b='bbb',c=3,d=4)
    assert dict(o1) == { 'a':1,'b':'bbb','c':3,'d':4 }



from __future__ import annotations

from json import JSONEncoder

from dataclasses import is_dataclass, asdict
from datetime import datetime
from typing import Callable, Union

DEFAULT_FMT = "%Y-%m-%d %H:%M:%S"
DatetimeFmt = Union[str, Callable[[datetime], str]]


class Encoder(JSONEncoder, Callable[[...], JSONEncoder]):
    """``JSONEncoder`` that handles ``datatime``s and ``dataclass``es by default.

    Pass to ``json.dump`` or ``json.dumps`` like:

    >>> from utz import dataclass, Encoder, fromtimestamp, json  # Convenience imports from standard library
    >>> epoch = fromtimestamp(0)
    >>> json.dumps({ 'epoch': epoch }, cls=Encoder)
    '{"epoch": "1969-12-31 19:00:00"}'
    >>> json.dumps({ 'epoch': epoch }, cls=Encoder("%Y-%m-%d"), indent=2)
    '''{
      "epoch": "1969-12-31"
    }'''
    >>> @dataclass
    >>> class A:
    >>>     n: int
    >>>
    >>> json.dumps(A(111), cls=Encoder)
    '{"n": 111}'
    """
    def __init__(
        self,
        dt_fmt: DatetimeFmt | None = DEFAULT_FMT,
        dataclasses: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.dt_fmt = dt_fmt
        self.dataclasses = dataclasses

    def default(self, o) -> "str":
        if self.dataclasses and is_dataclass(o):
            return asdict(o)
        elif isinstance(o, datetime):
            dt_fmt = self.dt_fmt or DEFAULT_FMT
            if callable(dt_fmt):
                return dt_fmt(o)
            elif isinstance(dt_fmt, str):
                return o.strftime(self.dt_fmt)
            else:
                raise ValueError(f"Unexpected {dt_fmt=}")
        else:
            return super().default(o)

    def __call__(self, *args, **kwargs) -> "Encoder":
        """Convenience wrapper, allowing an ``Encoder`` instance to be passed to e.g. ``json.dump(cls=...)``."""
        return Encoder(
            dt_fmt=self.dt_fmt,
            dataclasses=self.dataclasses,
            *args,
            **kwargs,
        )




from functools import cached_property
import pytest
from utz import use


# convenient way to verify that a name is undefined
nope = pytest.raises(NameError)


def test_dict():
    x={'a':11,'b':22}
    with use(x):
        assert a == 11
        assert b == 22
    with nope: a
    with nope: b


def test_locals_override():
    a = 'aaa'
    x={'a':11,'b':22}
    with use(x):
        # NOTE: existing local is not overridden; this is a limitation of locals() in function scope that I couldn't find a way around
        assert a == 'aaa'
        assert b == 22

    assert a == 'aaa'
    with nope: b

    with pytest.raises(RuntimeError):
        with use(x, local_conflict='err'):
            pass


TEST_GLOBAL = 'TEST GLOBAL'
def test_global_override_restore():
    a = 'aaa'
    x = { 'a':11, 'b':22, 'TEST_GLOBAL': 'overridden global' }
    with use(x):
        # NOTE: existing local is not overridden; this is a limitation of locals() in function scope that I couldn't find a way around
        assert a == 'aaa'
        assert b == 22
        assert TEST_GLOBAL == 'overridden global'

    assert a == 'aaa'
    with nope: b
    assert TEST_GLOBAL == 'TEST GLOBAL'


class Foo:
    def __init__(self, name):
        self.name = name

    STATIC_MEMBER = "STATIC MEMBER"

    @classmethod
    def cm1(cls): return f'class method 1'
    @classmethod
    def cm2(cls, n): return f'class method 2: {n}'

    @staticmethod
    def sm1(): return 'static method 1'
    @staticmethod
    def sm2(n): return f'static method 2: {n}'

    def method1(self): return 'method 1'
    def method2(self, n): return f'method 2: {n}'

    @property
    def name_reversed(self): return ''.join(reversed(self.name))

    @cached_property
    def name_2x(self): return self.name + self.name

    def __str__(self): return f'Foo({self.name})'


def test_method():
    from utz.methods import Methods
    m = Methods(Foo('abc'))
    assert m.ivars == {'name'}
    assert m.methods == {'method1','method2'}
    assert m.cmethods == {'cm1','cm2'}
    assert m.smethods == {'sm1','sm2'}
    assert m.smembers == {'STATIC_MEMBER'}
    assert m.properties == {'name_reversed'}
    assert m.cached_properties == {'name_2x'}


def test_cls_all():
    foo = Foo('abc')

    with use(foo):
        # everything from `foo` is in scope!
        assert method1() == 'method 1'
        assert method2(111) == 'method 2: 111'
        assert cm1() == 'class method 1'
        assert cm2(222) == 'class method 2: 222'
        assert sm1() == 'static method 1'
        assert sm2(222) == 'static method 2: 222'
        assert STATIC_MEMBER == 'STATIC MEMBER'
        assert name == 'abc'
        assert name_2x == 'abcabc'
        assert name_reversed == 'cba'

    # nothing from `foo` is in scope
    with nope: method1
    with nope: method2
    with nope: cm1
    with nope: cm2
    with nope: sm1
    with nope: sm2
    with nope: STATIC_MEMBER
    with nope: name
    with nope: name_2x
    with nope: name_reversed


def test_cls_ivars_only():
    foo = Foo('abc')

    with use(foo, include='ivars'):
        # only the instance var `name` has been brought into scope
        assert name == 'abc'
        with nope: method1
        with nope: method2
        with nope: cm1
        with nope: cm2
        with nope: sm1
        with nope: sm2
        with nope: STATIC_MEMBER
        with nope: name_2x
        with nope: name_reversed

    # out here it's gone
    with nope: name


def test_exclude_multiple_types():
    foo = Foo('abc')

    # choose one or more classes via `include`/`exclude`
    with use(foo, exclude=['cmethods','smethods']):
        # everything from `foo` is in scope!
        assert method1() == 'method 1'
        assert method2(111) == 'method 2: 111'
        assert STATIC_MEMBER == 'STATIC MEMBER'
        assert name == 'abc'
        assert name_2x == 'abcabc'
        assert name_reversed == 'cba'

        with nope: cm1
        with nope: cm2
        with nope: sm1
        with nope: sm2

    # nothing from `foo` is in scope
    with nope: method1
    with nope: method2
    with nope: STATIC_MEMBER
    with nope: name
    with nope: name_2x
    with nope: name_reversed



#!/usr/bin/env python
from __future__ import annotations

from json import loads
from sys import stderr

import subprocess
from functools import partial
from subprocess import check_call, CalledProcessError, CompletedProcess, DEVNULL, Popen, PIPE
from typing import Union

from .cmd import Cmd
from .log import Log, silent
from .util import Arg, Elides, parse_cmd

err = partial(print, file=stderr)


Json = Union[None, list, dict, str, int, float, bool]


def run(
    *args: Arg,
    dry_run: bool = False,
    log: Log = err,
    check: bool = True,
    **kwargs,
) -> CompletedProcess | None:
    """Convenience wrapper for ``subprocess.check_call``."""
    cmd = Cmd.mk(*args, **kwargs)
    if dry_run:
        if log:
            log(f'Would run: {cmd}')
    else:
        args, kwargs = cmd.compile(log=log)
        if check:
            check_call(args, **kwargs)
            return None
        else:
            return subprocess.run(args, **kwargs)


sh = run


def output(
    *args: Arg,
    dry_run: bool = False,
    both: bool = False,
    err_ok: bool | None = False,
    log: Log = err,
    **kwargs,
) -> bytes | None:
    """Convenience wrapper for ``subprocess.check_output``.

    By default, logs commands to `err` (stderr) before running (pass `log=None` to disable).

    If ``err_ok=True``, exceptions will be caught and suppressed, and the stdout up to that point will be returned.
    If ``err_ok=None``, exceptions will be caught and suppressed, and ``None`` will be returned.

    ``both=True`` is an alias for ``stderr=STDOUT``.
    """
    cmd = Cmd.mk(*args, **kwargs)
    if dry_run:
        if log:
            log(f'Would run: {cmd}')
        return None
    else:
        args, kwargs = cmd.compile(log=log, both=both)
        try:
            proc = Popen(args, stdout=PIPE, **kwargs)

            # Stream and capture the output in real-time
            output = b''
            for line in proc.stdout:
                output += line

            proc.wait()
            if proc.returncode != 0:
                raise CalledProcessError(proc.returncode, cmd, output=output)
            else:
                return output
        except CalledProcessError as e:
            if err_ok is True:
                return e.output
            elif err_ok is None:
                return None
            else:
                raise e


def text(*args, **kwargs) -> str | None:
    return output(*args, **kwargs).decode()


def json(
    *cmd: Arg,
    dry_run: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> Json:
    """Run a command, parse the output as JSON, and return the parsed object."""
    out = output(*cmd, dry_run=dry_run, err_ok=err_ok, **kwargs)
    if out is None or err_ok is True and not out:
        return None
    return loads(out.decode())


def check(
    *cmd: Arg,
    stdout=DEVNULL,
    stderr=DEVNULL,
    **kwargs,
):
    """Run a command, return True iff it runs successfully (i.e. exits with code 0)."""
    try:
        run(*cmd, stdout=stdout, stderr=stderr, **kwargs)
        return True
    except CalledProcessError:
        return False


def lines(
    *cmd: Arg,
    dry_run: bool = False,
    err_ok: bool = False,
    keep_trailing_newline: bool = False,
    **kwargs,
) -> list[str] | None:
    """Return the lines written to stdout by a command."""
    out = output(*cmd, dry_run=dry_run, err_ok=err_ok, **kwargs)
    if err_ok is None and out is None:
        return None

    lines = [
        line.rstrip('\n')
        for line in
        out.decode().split('\n')
    ]

    if not keep_trailing_newline and lines and not lines[-1]:
        lines = lines[:-1]

    return lines


def line(
    *cmd: Arg,
    empty_ok: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> str | None:
    """Run a command, verify that it returns a single line of output, and return that line."""
    _lines = lines(*cmd, err_ok=err_ok, **kwargs)
    if (empty_ok or err_ok is not False) and not _lines:
        return None
    elif len(_lines) == 1:
        return _lines[0]
    else:
        raise ValueError(f'Expected 1 line, found {len(_lines)}:\n\t%s' % '\n\t'.join(_lines))


from . import aio
from .pipeline import pipeline
from .named_pipes import named_pipes


# Omit "json", to avoid colliding with stdlib
__all__ = [
    'check',
    'err',
    'line',
    'lines',
    'named_pipes',
    'output',
    'pipeline',
    'run',
    'sh',
    'silent',
    'text',
    'Arg',
    'Cmd',
    'Elides',
    'Log',
]



import os
from pytest import fixture, raises, warns
import warnings
from typing import Generator

from utz.environ import env, OnConflict, OnExit, EnvPatchConflict


@fixture
def clean_env() -> Generator[None, None, None]:
    """Fixture to preserve and restore original environment."""
    original = dict(os.environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(original)


class TestEnvBasics:
    def test_singleton(self):
        """Test that Env is a proper singleton."""
        from utz.environ import Env
        assert env is Env()
        assert Env() is Env()

    def test_dict_interface(self, clean_env):
        """Test basic dictionary-like operations."""
        env["TEST_VAR"] = "value1"
        try:
            assert env["TEST_VAR"] == "value1"
            assert "TEST_VAR" in env
            assert env.get("TEST_VAR") == "value1"
            assert env.get("NONEXISTENT", "default") == "default"
        finally:
            del env["TEST_VAR"]
        assert "TEST_VAR" not in env

        # Test iteration
        env.update({"A": "1", "B": "2"})
        assert set(env.keys()) >= {"A", "B"}
        assert set(env.items()) >= {("A", "1"), ("B", "2")}
        assert set(env.values()) >= {"1", "2"}

    def test_type_coercion(self, clean_env):
        """Test that values are properly converted to strings."""
        env["INT"] = 42
        env["BOOL"] = True
        env["FLOAT"] = 3.14

        assert env["INT"] == "42"
        assert env["BOOL"] == "True"
        assert env["FLOAT"] == "3.14"


class TestContextManager:
    def test_basic_context(self, clean_env):
        """Test basic context manager functionality."""
        assert "TEST_VAR" not in env

        with env(TEST_VAR="temporary"):
            assert env["TEST_VAR"] == "temporary"

        assert "TEST_VAR" not in env

    def test_dict_syntax(self, clean_env):
        """Test dictionary syntax for context manager."""
        with env({"TEST_VAR": "value1", "TEST_VAR2": "value2"}):
            assert env["TEST_VAR"] == "value1"
            assert env["TEST_VAR2"] == "value2"

        assert "TEST_VAR" not in env
        assert "TEST_VAR2" not in env

    def test_kwargs_syntax(self, clean_env):
        """Test kwargs syntax for context manager."""
        with env(TEST_VAR="value1", TEST_VAR2="value2"):
            assert env["TEST_VAR"] == "value1"
            assert env["TEST_VAR2"] == "value2"

        assert "TEST_VAR" not in env
        assert "TEST_VAR2" not in env

    def test_mixed_syntax_error(self, clean_env):
        """Test that mixing dict and kwargs raises an error."""
        with raises(ValueError):
            with env({"TEST": "value"}, OTHER="value"):
                pass

    def test_existing_var_restore(self, clean_env):
        """Test that existing variables are properly restored."""
        env["EXISTING"] = "original"

        with env(EXISTING="temporary"):
            assert env["EXISTING"] == "temporary"

        assert env["EXISTING"] == "original"

    def test_multiple_nested_contexts(self, clean_env):
        """Test nested context managers."""
        with env(VAR1="outer1", VAR2="outer2"):
            assert env["VAR1"] == "outer1"
            assert env["VAR2"] == "outer2"

            with env(VAR1="inner1", VAR3="inner3"):
                assert env["VAR1"] == "inner1"
                assert env["VAR2"] == "outer2"
                assert env["VAR3"] == "inner3"

            assert env["VAR1"] == "outer1"
            assert env["VAR2"] == "outer2"
            assert "VAR3" not in env

        assert "VAR1" not in env
        assert "VAR2" not in env


class TestConflictHandling:
    def test_error_on_conflict(self, clean_env):
        """Test that conflicts raise an error by default."""
        with raises(EnvPatchConflict, match="TEST_VAR"):
            with env(TEST_VAR="original"):
                env["TEST_VAR"] = "modified"

        assert "TEST_VAR" not in env

    def test_warn_on_conflict(self, clean_env):
        """Test warning behavior for conflicts."""
        with warnings.catch_warnings(record=True) as w:
            with env(TEST_VAR="original", _on_conflict=OnConflict.WARN):
                env["TEST_VAR"] = "modified"

            assert len(w) == 1
            assert issubclass(w[0].category, RuntimeWarning)
            assert "TEST_VAR" in str(w[0].message)

    def test_ignore_conflict(self, clean_env):
        """Test ignore behavior for conflicts."""
        with env(TEST_VAR="original", _on_conflict=OnConflict.IGNORE):
            env["TEST_VAR"] = "modified"
            # Should not raise or warn

    def test_unmodified_always_reset(self, clean_env):
        """Test that unmodified variables are always reset regardless of on_exit."""
        with env(MODIFIED="value1", UNMODIFIED="value2", on_conflict=OnConflict.IGNORE, on_exit=OnExit.SKIP):
            env["MODIFIED"] = "changed"
            assert env["UNMODIFIED"] == "value2"

        assert "UNMODIFIED" not in env  # Should always be reset
        assert env["MODIFIED"] == "changed"

    def test_skip_conflicted(self, clean_env):
        """Test that conflicted variables can be skipped on exit."""
        with (
            warns(RuntimeWarning, match="TEST_VAR"),
            env(TEST_VAR="original", on_conflict=OnConflict.WARN, on_exit=OnExit.SKIP)
        ):
            env["TEST_VAR"] = "modified"

        assert env["TEST_VAR"] == "modified"

    def test_reset_conflicted(self, clean_env):
        """Test that conflicted variables can be reset on exit."""
        env["TEST_VAR"] = "initial"

        with env(TEST_VAR="original", on_conflict=OnConflict.IGNORE, on_exit=OnExit.RESET):
            env["TEST_VAR"] = "modified"

        assert env["TEST_VAR"] == "initial"



from __future__ import annotations

from io import UnsupportedOperation, StringIO
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
from typing import AnyStr, IO, Literal

from utz.process import Cmd


def pipeline(
    cmds: list[str] | list[list[str]] | list[Cmd],
    out: str | IO[AnyStr] | None = None,
    mode: Literal['b', 't', None] = None,
    wait: bool = True,
    both: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> str | list[Popen] | None:
    """Run a pipeline of commands, writing the final stdout to a file or ``IO``, or returning it as a ``str``"""
    processes = []
    prev_process: Popen | None = None

    cmds = [
        cmd if isinstance(cmd, Cmd) else
        Cmd.mk(cmd, **kwargs)
        for cmd in cmds
    ]

    return_output = False
    if out is None:
        out = StringIO()
        return_output = True
        if not wait:
            raise ValueError("Can't return output with `wait=False`")

    if mode is None:
        mode = 't' if isinstance(out, StringIO) else 'b'

    # If out is StringIO/BytesIO, use PIPE instead
    use_pipe = False
    if hasattr(out, 'write'):
        try:
            out.fileno()
        except UnsupportedOperation:
            use_pipe = True

    # Collect stderr for error reporting (if not redirecting stderr to stdout)
    stderr_pipes = [] if not both else None

    for i, cmd in enumerate(cmds):
        is_last = i + 1 == len(cmds)

        # For the first process, take input from the original source
        stdin = None if prev_process is None else prev_process.stdout

        def mkproc(stdout=PIPE):
            args, kwargs = cmd.compile(both=both)
            kwargs['stderr'] = STDOUT if both else PIPE
            return Popen(
                args,
                stdin=stdin,
                stdout=stdout,
                **kwargs,
            )

        # For the last process
        if is_last:
            if use_pipe:
                proc = mkproc()
                # Read the output and write to the StringIO/BytesIO
                output = proc.stdout.read()
                if mode == 't' and isinstance(output, bytes):
                    output = output.decode()
                out.write(output)
            else:
                with (open(out, f'w{mode}') if isinstance(out, str) else out) as pipe_fd:
                    proc = mkproc(pipe_fd)

        # For intermediate processes, output to a pipe
        else:
            proc = mkproc()

        if prev_process is not None:
            prev_process.stdout.close()

        processes.append(proc)
        if not both:
            stderr_pipes.append(proc.stderr)
        prev_process = proc

    if not wait:
        return processes

    for p in processes:
        p.wait()

    def get_output():
        if isinstance(out, str):
            with open(out, 'rt') as f:
                return f.read()
        else:
            return out.getvalue()

    if not err_ok:
        # Check for errors + `raise`
        for i, p in enumerate(processes):
            returncode = p.returncode

            if returncode != 0:
                # Collect stderr from the process, if available
                stdout_output = p.stdout.read()
                if isinstance(stdout_output, bytes):
                    stdout_output = stdout_output.decode('utf-8', errors='replace')

                if stderr_pipes:
                    stderr_output = stderr_pipes[i].read()
                elif p.stderr:
                    stderr_output = p.stderr.read()
                else:
                    stderr_output = None
                if isinstance(stderr_output, bytes):
                    stderr_output = stderr_output.decode('utf-8', errors='replace')

                # Prepare the original command for the error message
                cmd_args, _ = cmds[i].compile()
                cmd_str = ' '.join(str(arg) for arg in cmd_args) if isinstance(cmd_args, list) else cmd_args

                if not stdout_output and (not both or not stderr_output):
                    stdout_output = get_output()
                raise CalledProcessError(
                    returncode,
                    cmd_str,
                    output=stdout_output,
                    stderr=stderr_output,
                )

    if return_output:
        return get_output()



from unittest.mock import patch

from datetime import datetime
import pytest
from pytest import approx
from pytz import UTC

from utz import b62, b64, b90, now, o, today, Time

to_dt = now.to_dt


def test_now():
    fmt = '%Y-%m-%dT%H:%M:%SZ'
    assert fmt == now.FMTS.iso

    nw = now()
    assert str(nw) == datetime.now().astimezone(UTC).strftime(fmt)

    fmt = '%Y-%m-%d'
    assert fmt == today.FMTS.iso
    tdy = today()
    assert str(tdy) == datetime.now().astimezone(UTC).strftime(fmt)

    n1 = now()
    n2 = now()
    n3 = now()
    Δ1 = float(n2) - float(n1)
    Δ2 = float(n3) - float(n2)
    assert 0 <= Δ1 < .1, Δ1
    assert 0 <= Δ2 < .1, Δ2

    Δ1 = int(n2) - int(n1)
    Δ2 = int(n3) - int(n2)
    assert 0 <= Δ1 <= 1, Δ1
    assert 0 <= Δ2 <= 1, Δ2
    assert Δ1 == 0 or Δ2 == 0


@pytest.mark.parametrize("debug", [False])  # set to true to print the expected values for new test cases
@pytest.mark.parametrize(
    "codec,cases,t2021",
    (
        (
            b64,
            (
                    o(unit= 's', ch='+', len=6, first_until='2038-08-04T09:32:48Z', len_until='4182-03-14T03:28:00Z'),
                    o(unit='ms', ch='M', len=7, first_until='2026-08-27T02:19:40.480000Z', len_until='2111-08-01T07:19:33.184000Z'),
                    o(unit='us', ch='3', len=9, first_until='2032-07-30T09:31:33.747776Z', len_until='2549-11-30T12:09:02.965824Z'),
            ),
            ('+SuZKz', 'KNuCZyz', '2gvgesKyz'),
        ),
        (
            b62,
            (
                    o(unit= 's', ch='A', len=6, first_until='2028-07-15T14:30:34Z', len_until='3799-06-08T08:23:06Z'),
                    o(unit='ms', ch='d', len=7, first_until='2025-10-29T11:47:34.506000Z', len_until='2083-06-04T14:46:33.194000Z'),
                    o(unit='us', ch='G', len=9, first_until='2025-06-19T02:29:50.672362Z', len_until='2406-01-02T18:06:37.841642Z'),
            ),
            ('At4G0V', 'bTwXvhz', 'GWAZusO3r'),
        ),
        (
            b90,
            (
                    o(unit= 's', ch=':', len=5, first_until='2026-02-27T15:46:30Z', len_until='2159-03-22T15:46:30Z'),
                    o(unit='ms', ch='#', len=7, first_until='2037-07-20T20:40:47.190000Z', len_until='3502-09-11T10:10:47.190002Z'),
                    o(unit='us', ch='D', len=8, first_until='2026-02-04T20:12:22.247190Z', len_until='2107-12-11T09:33:22.247190Z'),
            ),
            ('8Od[z', '#"R^v[z', 'AZK=Tv[z'),
        ),
    ),
)
def test_encodings(debug, codec, cases, t2021):
    # For each unit (s,ms,us), serialize the current epoch time using `b64`, and verify a few properties:
    # - `ch`: first char of serialized string (which changes least frequently)
    # - `len`: serialized string length
    # - `first_until`: time when the first char will change
    # - `len_until`: time when the length will change
    last = codec.i2s[-1]
    for case in cases:
        unit = case.unit
        t = now()
        n = getattr(t, unit)
        s = codec(n)
        N = len(s)
        s2dt = getattr(now, f'from_{unit}')
        first_until = s2dt(codec(s[0] + last*(N-1)))
        len_until = s2dt(codec(last*N))

        if debug:
            unit_str = '% 4s' % f"'{unit}'"

            def fmt(d):
                if unit == 's':
                    return str(now(d=d))
                else:
                    return str(now(fmt='micro', d=d))

            print(f"o(unit={unit_str}, ch='{s[0]}', len={N}, first_until='{fmt(first_until)}', len_until='{fmt(len_until)}'),")
        else:
            assert s[0] == case.ch
            assert N == case.len
            assert first_until == to_dt(case.first_until)
            assert len_until == to_dt(case.len_until)

    instant = to_dt('20210101')
    assert (
        codec(now(d=instant). s),
        codec(now(d=instant).ms),
        codec(now(d=instant).us),
    ) == t2021


class FakeTimer:
    def __init__(self, start=0.0, increment=0.1):
        self.value = start
        self.increment = increment
        # self.orig_increment = increment

    def __call__(self):
        result = self.value
        self.value += self.increment
        # self.increment += self.orig_increment

        return result


def test_timer():
    with patch('utz.time.perf_counter', FakeTimer()):
        time = Time()
        time("a")
        time("b")
        time("c")
        time()

    assert time.times == approx({
        'a': 0.1,
        'b': 0.1,
        'c': 0.1,
    })


def test_timer_ctx():
    with patch('utz.time.perf_counter', FakeTimer()):
        time = Time()
        with time("a"):
            time("b")
            time("c")

        with time("d"):
            pass

        with time("e"), time("f"):
            with time("g"):
                time("h")
                time()

    assert time.times == approx({
        'a': 0.4,
        'b': 0.1,
        'c': 0.1,
        'd': 0.1,
        'e': 0.9,
        'f': 0.6,
        'g': 0.3,
        'h': 0.1,
    })



from os import getcwd
from tempfile import NamedTemporaryFile

from utz import docker, line, lines, process, run
from utz.docker.dsl import *


def test_docker_build_and_run():
    image_name = 'docker-dsl-test'
    file = docker.File()
    with file:
        FROM('bash')
        WORKDIR('/abc')
        ENTRYPOINT('pwd')
    image = file.build(image_name)
    name = f'{image_name}-{hash(image)}'
    image.run(name)
    try:
        assert line('docker', 'logs', name) == '/abc'
    finally:
        run('docker', 'container', 'rm', name)


def test_docker_file_contents():
    with NamedTemporaryFile(prefix='Dockerfile.', dir=getcwd()) as tmpfile:
        path = tmpfile.name
        file = docker.File(path)
        with file:
            FROM('bash')
            WORKDIR('/abc')
            NOTE('Set up some env vars')
            LN()
            LABEL(**{'a.b c': '1 \n2 \n3', 'aaa': 111})
            ENV('aaa=AAA', 'b=B=B', c='333')
            LABEL(r'a\\b=c\\d', bbb='BBB', **{r'e\f': 'g\\h'})
            ENTRYPOINT('a', 'b c', shell=True)
            ENTRYPOINT('pwd')

        assert file.lines == [
            'FROM bash',
            'WORKDIR /abc',
            '# Set up some env vars',
            '',
            'LABEL "a.b c"="1 \\\n2 \\\n3" "aaa"="111"',
            'ENV aaa=AAA b=B=B "c"="333"',
            r'LABEL a\\b=c\\d "bbb"="BBB" "e\\f"="g\\h"',
            "ENTRYPOINT a 'b c'",
            'ENTRYPOINT ["pwd"]',
        ]

        from datetime import datetime
        now = datetime.now().strftime('%Y%m%d_%H%M%S')
        img_name = f'docker-label-env-test-{now}'
        image = file.build(img_name)
        [img] = process.json('docker', 'inspect', img_name)
        assert img['Config']['Labels'] == {'a.b c': '1 2 3', 'aaa': '111', r'a\b': r'c\d', 'bbb': 'BBB', 'e\\f': 'g\\h'}
        container_name = f'{img_name}-{hash(image)}'
        image.run(container_name)
        try:
            assert line('docker', 'logs', container_name) == '/abc'
        finally:
            run('docker', 'container', 'rm', container_name)
            run('docker', 'image', 'rm', img_name)


def test_docker_demo():
    def build_python_git_image(python_version, tag):
        """Build a Docker image with a specified Python version, and `git` installed"""
        with docker.File(tag=tag):
            FROM(f'python:{python_version}-slim')
            RUN(
                'apt-get update',
                'apt-get install -y git',
            )
            ENTRYPOINT('python', '-V')

    [
        build_python_git_image(pv, f'python-git:{pv}')
        for pv in ('3.7.9', '3.8.6', '3.9.1',)
    ]

    assert lines('docker', 'run', 'python-git:3.9.1') == ['Python 3.9.1']




from os.path import exists, isdir, isfile, join
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from utz.zip import try_unzip

def test_unzip():
    files = {
        'a': 'yay',
        'b/c': 'CCC',
        'b/d': 'DDD',
    }

    with TemporaryDirectory() as tmpdir:
        name = 'test'
        zip_path = join(tmpdir, f'{name}.zip')
        with ZipFile(zip_path,'w') as z:
            for path, contents in files.items():
                z.writestr(path, contents)

        zip_dir = join(tmpdir, name)
        assert try_unzip(zip_path) == zip_dir
        assert not exists(zip_path)
        assert exists(zip_dir)
        assert isdir(zip_dir)

        for path, contents in files.items():
            with open(join(zip_dir,path),'r') as f:
                assert f.read() == contents

        a_path = join(zip_dir,'a')
        assert try_unzip(a_path) == a_path
        assert isfile(a_path)

    with TemporaryDirectory() as tmpdir:
        name = 'test'
        zip_path = join(tmpdir, name)
        with ZipFile(zip_path,'w') as z:
            for path, contents in files.items():
                z.writestr(path, contents)

        zip_dir = zip_path
        assert try_unzip(zip_path) == zip_dir
        assert not exists(f'{zip_path}.zip')
        assert exists(zip_dir)
        assert isdir(zip_dir)

        for path, contents in files.items():
            with open(join(zip_dir,path),'r') as f:
                assert f.read() == contents

        a_path = join(zip_dir,'a')
        assert try_unzip(a_path) == a_path
        assert isfile(a_path)



from os import getcwd
from os.path import basename, join, exists
from tempfile import TemporaryDirectory

from utz import cd, cd_tmpdir

cwd = getcwd()


def test_cd():
    with TemporaryDirectory(dir='.') as subdir:
        name = basename(subdir)
        with cd(name):
            assert getcwd() == join(cwd, name)
            with cd('..'):
                assert getcwd() == cwd
        assert getcwd() == cwd
    assert not exists(subdir)


def test_cd_tmpdir():
    with cd_tmpdir(dir='.') as subdir:
        name = basename(subdir)
        assert getcwd() == join(cwd, name)
    assert not exists(subdir)

    with cd_tmpdir(dir='abc') as subdir:
        name = basename(subdir)
        assert getcwd() == join(cwd, 'abc', name)
    assert not exists(subdir)



from __future__ import annotations

import os
import enum
import warnings
from collections.abc import MutableMapping
from contextlib import contextmanager
from typing import Any, Iterator, overload


class OnConflict(enum.Enum):
    """How to handle variables that were changed during the context."""
    ERROR = "error"
    WARN = "warn"
    IGNORE = "ignore"


class OnExit(enum.Enum):
    """What to do with conflicted variables on context exit."""
    RESET = "reset"  # Reset conflicted vars to original value if it existed, otherwise remove
    SKIP = "skip"   # Leave conflicted values in place


class EnvPatchConflict(Exception):
    """Raised when environment variables were modified during the context."""
    def __init__(self, conflicts: dict[str, tuple[str, str]]):
        self.conflicts = conflicts
        items = [f"{k}: expected={v[0]!r}, found={v[1]!r}"
                 for k, v in conflicts.items()]
        super().__init__(
            "Environment variables were modified during the context:\n  " +
            "\n  ".join(items)
        )


class Env(MutableMapping):
    """
    A singleton wrapper around os.environ that provides dictionary-like access to environment
    variables and supports temporarily modifying values using a context manager.

    Usage:
        # Dictionary-like access
        env["MY_VAR"] = "value1"
        print(env["MY_VAR"])
        print(env.get("NON_EXISTENT", "default"))

        # Context manager usage with conflict handling
        with env({"MY_VAR": "temporary"},
                on_conflict=OnConflict.ERROR,
                on_exit=OnExit.RESET):
            print(env["MY_VAR"])  # temporary

        # Or with kwargs syntax
        with env(MY_VAR="temporary",
                _on_conflict=OnConflict.WARN,
                _on_exit=OnExit.SKIP):
            print(env["MY_VAR"])  # temporary
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        # Inherit all methods/attributes from os.environ that we haven't overridden
        self.__dict__.update(os.environ.__dict__)

    def __getitem__(self, key: str) -> str:
        return os.environ[key]

    def __setitem__(self, key: str, value: str) -> None:
        os.environ[key] = str(value)

    def __delitem__(self, key: str) -> None:
        del os.environ[key]

    def __iter__(self) -> Iterator[str]:
        return iter(os.environ)

    def __len__(self) -> int:
        return len(os.environ)

    @contextmanager
    def _patch(
        self,
        updates: dict[str, str],
        on_conflict: OnConflict = OnConflict.ERROR,
        on_exit: OnExit = OnExit.RESET,
    ):
        """Internal method for handling environment variable patching."""
        # Convert all values to strings and store originals
        updates = {k: str(v) for k, v in updates.items()}
        original_values = {
            k: os.environ.get(k) for k in updates
        }

        # Apply updates
        os.environ.update(updates)

        try:
            yield self
        finally:
            # Check for modifications during context
            conflicts = {}
            for key, expected in updates.items():
                current = os.environ.get(key)
                if current != expected:
                    conflicts[key] = (expected, current)

            # Always reset unmodified variables
            unmodified = set(updates) - set(conflicts)
            for key in unmodified:
                orig = original_values[key]
                if orig is None:
                    if key in os.environ:
                        del os.environ[key]
                else:
                    os.environ[key] = orig

            # Handle conflicted variables based on on_exit policy
            if conflicts and on_exit == OnExit.RESET:
                for key in conflicts:
                    orig = original_values[key]
                    if orig is None:
                        if key in os.environ:
                            del os.environ[key]
                    else:
                        os.environ[key] = orig

            if conflicts:
                if on_conflict == OnConflict.ERROR:
                    raise EnvPatchConflict(conflicts)
                elif on_conflict == OnConflict.WARN:
                    warnings.warn(
                        f"Environment variables were modified during the context: {conflicts}",
                        RuntimeWarning
                    )

    @overload
    def __call__(
        self,
        updates: dict[str, str],
        *,
        on_conflict: OnConflict = OnConflict.ERROR,
        on_exit: OnExit = OnExit.RESET,
    ) -> Any: ...

    @overload
    def __call__(
        self,
        **kwargs: str | OnConflict | OnExit,
    ) -> Any: ...

    def __call__(
        self,
        updates: dict[str, str] | None = None,
        *,
        on_conflict: OnConflict = OnConflict.ERROR,
        on_exit: OnExit = OnExit.RESET,
        **kwargs,
    ) -> Any:
        """
        Support both dict and kwargs syntax for the context manager:
        with env({"KEY": "value"}) and with env(KEY="value")

        Special kwargs when using kwargs syntax:
            _on_conflict: OnConflict = OnConflict.ERROR
            _on_exit: OnExit = OnExit.RESET
        """
        if updates is not None and kwargs:
            raise ValueError("Cannot specify both dictionary and keyword arguments")

        if updates is not None:
            return self._patch(updates, on_conflict=on_conflict, on_exit=on_exit)

        # Extract special kwargs
        _on_conflict = kwargs.pop('_on_conflict', on_conflict)
        _on_exit = kwargs.pop('_on_exit', on_exit)

        return self._patch(kwargs, on_conflict=_on_conflict, on_exit=_on_exit)


# Create singleton instance
env = Env()



from __future__ import annotations

from contextlib import contextmanager

from typing import KeysView, ItemsView, ValuesView, Callable

from datetime import datetime as dt, timezone
from sys import stderr
from time import perf_counter
from types import TracebackType

from utz.proc import err
from utz.o import o


utc = timezone.utc


class Time:
    """Simple "timer" class.

    >>> from time import sleep
    >>> time = Time()
    >>> time("step 1")
    >>> sleep(1)
    >>> time("step 2")
    >>> sleep(1)
    >>> time()  # "close" "step 2"
    >>> print(f'Step 1 took {time["step 1"]:.1f}s, step 2 took {time["step 2"]:.1f}s.'

    Can also be used as a contextmanager:

    >>> time = Time()
    >>> with time("run"):
    >>>     sleep(1)
    >>> print(f'Run took {time["run"]:.1f}s')
    """
    def __init__(self, log: str | True | Callable[[str, float], str] | None = None):
        self.times = {}
        self._cur_timer = None
        self._cur_start = 0
        # `__{enter,exit}__`s push and pop timer names to and from this stack
        self._ctx_stack = []

        if log is True:
            log = "{k} took {v:.3g}s"
        if isinstance(log, str):
            def fmt_fn(k, v):
                return log.format(k=k, v=v)
            self._log = fmt_fn
        elif log is None:
            self._log = None
        else:
            self._log = log

    def save(self, name: str, duration: float):
        """Record a duration associated with a timer name, and optionally log a message about it."""
        self.times[name] = duration
        if self._log:
            err(self._log(name, duration))

    def end(self, ts: float | None = None):
        """End and save any active timer."""
        if self._cur_timer:
            prev_end = ts or perf_counter()
            duration = prev_end - self._cur_start
            self.save(self._cur_timer, duration)
            self._cur_timer = None
            self._cur_start = 0

    def __call__(self, name: str | None = None) -> "Time":
        """Primary method for recording a named timer."""
        self.end()  # End any active timer
        if name:
            # Start a new timer
            self._cur_timer = name
            self._cur_start = perf_counter()
        return self

    @contextmanager
    def ctx(self, name: str):
        with self(name):
            yield

    def __enter__(self):
        """Push the current timer onto the "context stack"."""
        if self._cur_timer is not None:
            self._ctx_stack.append((self._cur_timer, perf_counter()))
            self._cur_timer = None
            self._cur_start = 0
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        """End the most-recently `__enter__`d timer."""
        end = perf_counter()
        self.end(end)
        if self._ctx_stack:
            name, start = self._ctx_stack.pop()
            duration = end - start
            self.save(name, duration)

        if exc_value:
            raise exc_value

    def fmt(self, fmt_spec: str = ".3g") -> dict[str, float]:
        """Format the accumulated timings."""
        fmt_str = "{:%s}" % fmt_spec
        return {
            k: float(fmt_str.format(v))
            for k, v in self.times.items()
        }

    def __getitem__(self, name: str) -> float:
        return self.times[name]

    def keys(self) -> KeysView[str]:
        return self.times.keys()

    def values(self) -> ValuesView[float]:
        return self.times.values()

    def items(self) -> ItemsView[str, float]:
        return self.times.items()


class now:
    try:
        from pytz import UTC
        pytz = True
        def tz(d): return d.astimezone(now.UTC)
    except ImportError:
        pytz = False
        UTC = None
        def tz(d): return d

    EPOCH = dt(1970, 1, 1).replace(tzinfo=UTC)
    FMTS = o(
        iso='%Y-%m-%dT%H:%M:%SZ',
        short='%Y%m%dT%H%M%SZ',
        micro='%Y-%m-%dT%H:%M:%S.%fZ',
    )

    @property
    def fmts(self):
        return self.FMTS

    def __init__(self, fmt=None, d=None):
        FMTS = self.FMTS
        if fmt is None:
            fmt = FMTS.iso
        elif isinstance(fmt, str):
            if hasattr(FMTS, fmt):
                fmt = getattr(FMTS, fmt)
        if d is None:
            self.time = dt.now()
        elif isinstance(d, str):
            from dateutil.parser import parse
            self.time = parse(d)
        elif isinstance(d, dt):
            self.time = d
        if now.pytz:
            self.time = self.time.astimezone(now.UTC)
        else:
            # timezone will be local, not UTC; don't imply UTC in format
            stderr.write('pytz not installed; now()/today() will be based on OS TZ, which may not be UTC\n')
            assert fmt[-1] == 'Z'
            fmt = fmt[:-1]

        self.fmt = fmt

    def __call__(self, fmt): return self.time.strftime(fmt)
    def __getattr__(self, k): return getattr(self.time, k)
    def __str__(self): return self.time.strftime(self.fmt)
    def __repr__(self): return str(self)
    def __int__(self): return int((self.time - now.tz(now.EPOCH)).total_seconds())
    def __float__(self): return (self.time - now.tz(now.EPOCH)).total_seconds()

    @property
    def s(self): return int(self)

    @property
    def ds(self): return self.ms // 100

    @property
    def cs(self): return self.ms // 10

    @property
    def ms(self): return int(float(self) * 1e3)

    @property
    def μs(self): return int(float(self) * 1e6)

    @property
    def us(self): return int(float(self) * 1e6)

    @staticmethod
    def to_dt(s):
        from dateutil.parser import parse
        parsed = parse(s)
        if not parsed.tzinfo:
            parsed = parsed.replace(tzinfo=now.UTC)
        return parsed

    @staticmethod
    def fromtimestamp(s, tz=UTC): return dt.fromtimestamp(s, tz=tz)

    @staticmethod
    def from_s(s): return now.fromtimestamp(s)

    @staticmethod
    def from_ms(ms): return now.fromtimestamp(ms / 1e3)

    @staticmethod
    def from_μs(us): return now.fromtimestamp(us / 1e6)

    @staticmethod
    def from_us(us): return dt.fromtimestamp(us / 1e6, tz=now.UTC)


class today(now):
    FMTS = o(
        iso='%Y-%m-%d',
        short='%Y%m%d',
    )



from __future__ import annotations

from os.path import expandvars, expanduser
from subprocess import STDOUT

import shlex
from abc import ABC
from dataclasses import dataclass
from typing import Any, Sequence, Union

from utz.process.log import Log
from utz.process.util import Arg, Elides, flatten, ELIDED

Args = Union[str, list[str]]
Kwargs = dict[str, Any]
Compiled = tuple[Args, Kwargs]


class Cmd(ABC):
    """Interface for shell (`str`) and non-shell (`list[str]`) commands."""

    def compile(
        self,
        log: Log = None,
        both: bool = False,
    ) -> Compiled:
        """Compile a ``Cmd`` into ``args`` and ``kwargs`` suitable for passing to ``subprocess`` functions."""
        if log:
            log(f'Running: {self}')
        args, kwargs = self._compile()
        if both:
            if 'stderr' in kwargs and kwargs['stderr'] is not STDOUT:
                raise ValueError(f"`both=True` conflicts with `stderr={kwargs['stderr']}`")
            kwargs['stderr'] = STDOUT
        return args, kwargs

    def _compile(self):
        raise NotImplementedError

    @staticmethod
    def elide_cmd(cmd: str, elide: Elides) -> str:
        if elide:
            if isinstance(elide, str):
                elide = [elide]
            for s in elide:
                cmd = cmd.replace(s, ELIDED)
        return cmd

    def mk(
        *args: tuple[Arg, ...],
        sh: bool | str | None = None,
        shell: bool | str | None = None,
        executable: str | None = None,
        expanduser: bool | None = None,
        expandvars: bool | None = None,
        elide: Elides = None,
        **kwargs,
    ):
        """Construct a ``Cmd`` instance.

        - Handles `shell` and non-`shell` modes (including `shell` executable strings, with `sh` as a convenience alias
          for `shell`)
        - Optional strings to elide (e.g. secrets); **this has not been security-reviewed, and you should audit the
          relevant code paths if relying on it**.
        """
        if sh is not None:
            if shell is not None and sh != shell:
                raise ValueError(f"{sh=} != {shell=}")
            shell = sh
        if isinstance(shell, str):
            if executable and executable != shell:
                raise ValueError(f"{shell=} != {executable=}")
            executable = shell
            shell = True

        if len(args) == 1 and isinstance(args[0], str):
            if shell is True or shell is None:
                if expanduser or expandvars:
                    raise ValueError("Can't `expand{user,vars}` in shell mode")
                return ShellCmd(
                    args[0],
                    executable=executable,
                    elide=elide,
                    kwargs=kwargs,
                )
            else:
                raise ValueError("Expected `list[str]` command in non-shell mode")
        else:
            if shell is False or shell is None:
                return ArrayCmd(
                    args,
                    expanduser=expanduser,
                    expandvars=expandvars,
                    elide=elide,
                    kwargs=kwargs,
                )
            else:
                raise ValueError("Expected `str` command in shell mode")


@dataclass
class ShellCmd(Cmd):
    cmd: str
    executable: str | None = None
    elide: Elides = None
    kwargs: dict[str, Any] = None

    def __str__(self) -> str:
        return self.elide_cmd(self.cmd, self.elide)

    def _compile(self) -> tuple[str, Kwargs]:
        cmd = self.cmd
        if self.elide:
            if isinstance(self.elide, str):
                self.elide = [self.elide]
            for s in self.elide:
                cmd = cmd.replace(s, '****')

        return cmd, { **(self.kwargs or {}), 'shell': True, 'executable': self.executable, }


@dataclass
class ArrayCmd(Cmd):
    args: Sequence[Arg]
    expanduser: bool = False
    expandvars: bool = False
    elide: Elides = None
    kwargs: dict[str, Any] = None

    @property
    def cmd(self) -> list[str]:
        return [
            str(arg)
            for arg in flatten(self.args)
            if arg is not None
        ]

    def __str__(self) -> str:
        return self.elide_cmd(shlex.join(self.cmd), self.elide)

    def _compile(self) -> tuple[list[str], Kwargs]:
        cmd = self.cmd
        if self.expanduser:
            cmd = [ expanduser(arg) for arg in cmd ]
        if self.expandvars:
            cmd = [ expandvars(arg) for arg in cmd ]

        return cmd, { **(self.kwargs or {}), 'shell': False, }



from __future__ import annotations

from os import path

import shlex
from typing import Sequence, Union

Arg = Union[None, str, int, list['Arg'], tuple['Arg']]
ELIDED = '****'
Elides = Union[None, str, list[str]]


def flatten(args: Sequence[Arg]) -> tuple:
    """Recursively flatten a nested list of arguments."""
    if isinstance(args, list) or isinstance(args, tuple):
        return tuple(
            a
            for arg in args
            for a in flatten(arg)
        )
    else:
        return (args,)


def mk_cmd_str(cmd: Sequence[str] | str, elide: Elides = None) -> str:
    """Convert a command to a string, optionally eliding known sensitive values."""
    cmd_str = cmd if isinstance(cmd, str) else shlex.join(cmd)
    if elide:
        if isinstance(elide, str):
            elide = [elide]
        for s in elide:
            cmd_str = cmd_str.replace(s, ELIDED)
    return cmd_str


def parse_cmd(
    cmd: Sequence[Arg] | str,
    shell: bool | None = None,
    expanduser: bool | None = None,
    expandvars: bool | None = None,
    elide: Elides = None,
) -> tuple[list[str] | str, str]:
    """Flatten and stringify a command."""
    if shell:
        if not isinstance(cmd, str):
            raise ValueError("Expected string command in shell mode")
        if expanduser:
            cmd = path.expanduser(cmd)
        if expandvars:
            cmd = path.expandvars(cmd)
    else:
        cmd = [
            str(arg)
            for arg in flatten(cmd)
            if arg is not None
        ]
        if expanduser:
            cmd = [ path.expanduser(arg) for arg in cmd ]
        if expandvars:
            cmd = [ path.expandvars(arg) for arg in cmd ]

    cmd_str = mk_cmd_str(cmd, elide)
    return cmd, cmd_str



from contextlib import contextmanager
import utz.git.diff as diff
from utz.git.head import sha
from utz.git.repo import git_repo
from utz.o import o
from utz.proc import line, run


@contextmanager
def txn(start=None, msg=None, add=None,):
    """Contextmanager that creates a synthetic merge with __enter__- and __exit__-time HEADs as parents, and tree (and
    message, by default) of the latter.

    The yielded value is an o() dict-wrapper whose `msg` and `add` attrs will be used to configure a final commit
    """
    start = start or sha()
    ctx = o()
    if add:
        ctx.add = add
    if msg:
        ctx.msg = msg
    yield ctx

    if ctx('exit'):
        return

    end = sha()

    add = ctx('add')
    msg = ctx('msg') or f'merge txn: {start}, {end}'

    if add:
        run('git', 'add', add)

    if diff.exists(untracked=False, unstaged=False):
        run('git', 'commit', '-m', msg)
    elif ctx('empty'):
        run('git', 'commit', '--allow-empty', '-m', msg)

    if start != end:
        repo = git_repo()
        tree = repo.tree().hexsha
        head = line('git', 'commit-tree', tree, '-p', start, '-p', end, '-m', msg)
        run('git', 'reset', head)



from __future__ import annotations

from inspect import getfullargspec
from typing import Callable, Sequence

Deco = Callable[[Callable], Callable]


def decos(*args: Deco | Sequence[Deco]):
    """Compose decorators."""
    decos = [
        deco
        for decos in args
        for deco in (decos if isinstance(decos, Sequence) else [decos])
    ]

    def _fn(fn):
        for deco in reversed(decos):
            fn = deco(fn)
        return fn

    return _fn


def args(fn, kwargs):
    """Filter kwargs to match function signature."""
    return { k: v for k, v in kwargs.items() if recvs(fn, k) }


def call(fn, *_args, **kwargs):
    """Call a function with only the kwargs that it is able to receive."""
    return fn(*_args, **args(fn, kwargs))


def recvs(fn: Callable, k: str) -> bool:
    """True if ``fn`` takes a kwarg ``k``, or a var-kwargs and a wrapped descendant function does."""
    spec = getfullargspec(fn)
    if k in spec.args or k in spec.kwonlyargs:
        return True
    if spec.varkw:
        wrapped = getattr(fn, '__wrapped__', None)
        if wrapped:
            return recvs(wrapped, k)
    return False



from __future__ import annotations

def iec(
    value: float | str,
    binary: bool = True,
    gnu: bool = False,
    format: str = "%.3g",
) -> str:
    from humanize import naturalsize
    return naturalsize(
        value=value,
        binary=binary,
        gnu=gnu,
        format=format,
    )



from __future__ import annotations

import datetime
from dataclasses import dataclass
from datetime import datetime as dt, date
import re
from math import ceil
from typing import Union

from utz import Yield

# Types that can be passed to the Month constructor
Monthy = Union['YM', str, int, None]


@dataclass(init=False, order=True, eq=True, unsafe_hash=True)
class YM:
    y: int
    m: int

    RGX = re.compile(r'(?P<year>\d{4})(?:-?(?P<month>\d\d))?')

    def _init_from_str(self, arg):
        m = self.RGX.fullmatch(arg)
        if not m:
            raise ValueError('Invalid month string: %s' % arg)
        year = int(m['year'])
        month = int(m['month']) if m['month'] else 1
        if month > 12:
            raise ValueError(f"Invalid month {month} ({arg})")
        self.__init__(year, month)

    def _verify(self):
        if not isinstance(self.y, int):
            raise ValueError('Year %s must be int, not %s' % (str(self.y), type(self.y)))
        if not isinstance(self.m, int):
            raise ValueError('Month %s must be int, not %s' % (str(self.m), type(self.m)))

    def _init_now(self):
        now = dt.now()
        self.y = now.year
        self.m = now.month

    def __init__(self, *args, **kwargs):
        if kwargs:
            if args:
                raise ValueError(f'Pass args xor kwargs: {args}, {kwargs}')
            keys = list(kwargs.keys())
            if keys == ['y', 'm']:
                self.y = kwargs['y']
                self.m = kwargs['m']
            else:
                raise ValueError(f"Unrecognized kwargs: {kwargs}")
        elif len(args) == 2:
            self.y, self.m = int(args[0]), int(args[1])
            self._verify()
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                self._init_from_str(arg)
            elif isinstance(arg, int):
                self._init_from_str(str(arg))
            elif hasattr(arg, 'year') and hasattr(arg, 'month'):
                self.y = arg.year
                self.m = arg.month
                self._verify()
            elif arg is None:
                self._init_now()
            elif 'year' in arg and 'month' in arg:
                self.y = int(arg['year'])
                self.m = int(arg['month'])
                self._verify()
            else:
                raise ValueError('Unrecognized argument: %s' % str(arg))
        elif not args:
            self._init_now()
        else:
            raise ValueError('Unrecognized arguments: %s' % str(args))

    @property
    def year(self):
        return self.y

    @property
    def month(self):
        return self.m

    def str(self, sep=''):
        return '%d%s%02d' % (self.y, sep, self.m)

    def __str__(self):
        return self.str()

    def __int__(self):
        return int(str(self))

    def format(self, url, **kwargs):
        return url.format(ym=str(self), y=str(self.y), m=str(self.m), **kwargs)

    @property
    def dt(self):
        import pandas as pd
        return pd.to_datetime('%d-%02d' % (self.y, self.m))

    @property
    def date(self) -> date:
        return self.dt.date()

    @property
    def dates(self) -> tuple[datetime.date, datetime.date]:
        start = self.date
        end = (self + 1).date
        return start, end

    def __add__(self, n: int) -> 'YM':
        if not isinstance(n, int):
            raise ValueError('%s: can only add an integer to a Month, not %s: %s' % (str(self), str(type(n)), str(n)))
        y, m = self.y, self.m + n - 1
        y += m // 12
        m = (m % 12) + 1
        return YM(y, m)

    def __sub__(self, r: 'YM' | int) -> 'YM' | int:
        if isinstance(r, YM):
            y, m = self.y - r.y, self.m - r.m
            if m < 0:
                y -= 1
                m += 12
            return y * 12 + m
        elif isinstance(r, int):
            y, m = self.y, self.m - r - 1
            if m <= 0:
                years = int(ceil(-m / 12))
                y -= years
                m += 12 * years
                assert 0 <= m < 12
            m += 1
            return YM(y, m)
        else:
            raise ValueError(f'{self}: can only add subtract an int or YM from YM, not {r} ({type(r)})')

    def until(self, end: 'YM' = None, step: int = 1) -> Yield['YM']:
        cur: YM = YM(self)
        while end is None \
                or (step > 0 and cur < end) \
                or (step < 0 and cur > end):
            yield cur
            cur = cur + step

    def to(self, end: 'YM' = None, step: int = 1) -> Yield['YM']:
        return self.until(end=end + 1 if end else end, step=step)



from utz import dataclass, Encoder, json, fromtimestamp
from utz.time import utc


def test_datetimes():
    epoch = fromtimestamp(0, tz=utc)
    assert json.dumps({ 'epoch': epoch}, cls=Encoder) == '{"epoch": "1970-01-01 00:00:00"}'
    assert json.dumps({ 'epoch': epoch}, cls=Encoder, indent=2) == '\n'.join([
        '{',
        '  "epoch": "1970-01-01 00:00:00"',
        '}',
    ])
    assert json.dumps({ 'epoch': epoch}, cls=Encoder("%Y-%m-%d")) == '{"epoch": "1970-01-01"}'
    assert json.dumps({ 'epoch': epoch}, cls=Encoder("%Y-%m-%d"), indent=2) == '\n'.join([
        '{',
        '  "epoch": "1970-01-01"',
        '}',
    ])


@dataclass
class A:
    n: int


@dataclass
class B:
    arr: list[A]


def test_dataclasses():
    assert json.dumps(A(111), cls=Encoder) == '{"n": 111}'
    b = B([A(111), A(222)])
    b_str = '{"arr": [{"n": 111}, {"n": 222}]}'
    assert json.dumps(b, cls=Encoder) == b_str
    assert json.dumps({ 'b': b }, cls=Encoder) == '{"b": %s}' % b_str



# # Context Manager utilities

from __future__ import annotations

from asyncio import gather

from contextlib import AbstractContextManager, contextmanager, asynccontextmanager, AbstractAsyncContextManager, \
    AsyncExitStack
from typing import AsyncContextManager, ContextManager, Generator, Sequence, TypeVar, Iterator, Union

T = TypeVar("T")

# Shorthand for common return type of ``@contextmanager``-decorated functions
Yield = Generator[T, None, None]


def contexts(*ctxs: ContextManager | Sequence[ContextManager] | Iterator) -> ContextManager:
    """Compose context managers."""
    ctxs = [
        ctx
        for _ctxs in ctxs
        for ctx in (
            _ctxs
            if isinstance(_ctxs, Sequence)
            else [_ctxs]
        )
    ]

    @contextmanager
    def fn(ctxs):
        if not ctxs:
            yield
        else:
            [ ctx, *rest ] = ctxs
            with ctx as v, fn(rest) as vs:
                yield [ v, *(vs if vs else []) ]

    return fn(ctxs)


ctxs = contexts


async def aenter_ctx(ctx):
    """Enter an async context manager or adapt a sync one."""
    if hasattr(ctx, '__aenter__'):
        return await ctx.__aenter__()
    elif hasattr(ctx, '__enter__'):
        return ctx.__enter__()
    else:
        raise TypeError(f"Object {ctx} is not a context manager")

async def aexit_ctx(ctx, exc_type, exc_val, exc_tb):
    """Exit an async context manager or adapt a sync one."""
    if hasattr(ctx, '__aexit__'):
        return await ctx.__aexit__(exc_type, exc_val, exc_tb)
    elif hasattr(ctx, '__exit__'):
        return ctx.__exit__(exc_type, exc_val, exc_tb)
    else:
        raise TypeError(f"Object {ctx} is not a context manager")


CtxMgr = Union[ContextManager[T], AsyncContextManager[T]]


def acontexts(*ctxs: CtxMgr | Sequence[CtxMgr] | Iterator) -> AsyncContextManager:
    """Compose context managers, running their __aenter__ and __aexit__ methods concurrently."""
    # Flatten the list of context managers
    flat_ctxs = []
    for _ctxs in ctxs:
        if isinstance(_ctxs, Sequence):
            flat_ctxs.extend(_ctxs)
        else:
            flat_ctxs.append(_ctxs)

    @asynccontextmanager
    async def async_composed():
        exit_stack = []
        exc_details = (None, None, None)
        try:
            # Run all __aenter__ methods concurrently
            values = await gather(*(aenter_ctx(ctx) for ctx in flat_ctxs))
            exit_stack = list(flat_ctxs)  # Track contexts that need to be exited
            yield values
        except BaseException as e:
            # Store exception details for __aexit__ methods
            exc_details = (type(e), e, e.__traceback__)
            raise
        finally:
            # Run all __aexit__ methods concurrently
            if exit_stack:
                exit_results = await gather(
                    *(aexit_ctx(ctx, *exc_details) for ctx in reversed(exit_stack)),
                    return_exceptions=True
                )
                # Re-raise any exceptions from __aexit__ if no other exception is being propagated
                if exc_details[0] is None:
                    for result in exit_results:
                        if isinstance(result, BaseException):
                            raise result

    return async_composed()


actxs = acontexts
asyncnullcontext = AsyncExitStack  # https://stackoverflow.com/a/61488069


class WithExitHook(ContextManager[T], AsyncContextManager[T]):
    def __init__(self, ctx: CtxMgr[T], exit_hook: CtxMgr):
        self.ctx = ctx
        if not isinstance(exit_hook, (AbstractContextManager, AbstractAsyncContextManager)):
            raise ValueError("exit_hook must be an AbstractContextManager or AbstractAsyncContextManager")

        self.exit_hook = exit_hook

    def __enter__(self) -> T:
        return self.ctx.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.exit_hook:
            return self.ctx.__exit__(exc_type, exc_val, exc_tb)

    def __aenter__(self) -> T:
        return self.ctx.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if isinstance(self.exit_hook, AbstractContextManager):
            with self.exit_hook:
                return await self.ctx.__aexit__(exc_type, exc_val, exc_tb)
        else:
            async with self.exit_hook:
                return await self.ctx.__aexit__(exc_type, exc_val, exc_tb)


with_exit_hook = WithExitHook


class catch(AbstractContextManager):
    """``contextmanager`` that catches+verifies ``Exception``s"""
    def __init__(self, *excs):
        self.excs = excs

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            if len(self.excs) == 1:
                raise AssertionError(f'No {self.excs[0].__name__} was thrown')
            else:
                raise AssertionError(f'None of {",".join([e.__name__ for e in self.excs])} were thrown')
        
        if not [ isinstance(exc_value, exc) for exc in self.excs ]:
            raise exc_value

        return True


# ## `no`: context manager for verifying `NameError`s (undefined variable names)
no = catch(NameError)



from utz import deterministic_gzip_open, hash_file


def test_deterministic_gzip_write():
    for i in range(2):
        with deterministic_gzip_open('a.gz', 'w') as f:
            f.write('\n'.join(map(str, range(10))))

        assert hash_file('a.gz') == "dfbe03625c539cbc2a2331d806cc48652dd3e1f52fe187ac2f3420dbfb320504"



from __future__ import annotations

from contextlib import contextmanager, nullcontext
import os
from os import makedirs, mkdir, rmdir
from os.path import exists, join, sep, dirname
from tempfile import TemporaryDirectory


@contextmanager
def named_tmpdir(name: str = None, **kwargs):
    """``contextmanager`` for creating a temporary directory with an optional ``name``"""
    if name:
        if 'prefix' in kwargs:
            raise ValueError('`prefix` is not supported; use `name` instead')
        if 'suffix' in kwargs:
            raise ValueError('`suffix` is not supported; use `name` instead')
        with TemporaryDirectory(**kwargs) as tmpdir:
            dir = join(tmpdir, name)
            mkdir(dir)
            yield dir
    else:
        with TemporaryDirectory(**kwargs) as td:
            yield td


@contextmanager
def tmp_ensure_dir(path: str):
    """Ensure a directory (and its parents) exist; any dirs created are cleaned up on exit (expected to be empty)."""
    pcs = path.split(sep)
    cur = ''
    rms = []
    for pc in pcs:
        cur = os.path.join(cur, pc) if cur else pc
        if not exists(cur):
            makedirs(cur)
            rms.append(cur)
    try:
        yield
    finally:
        while rms:
            rm = rms.pop()
            rmdir(rm)


@contextmanager
def TmpDir(
    name: str | None = None,
    dir: str | None = None,
    **kwargs,
):
    """``TemporaryDirectory`` wrapper that temporarily creates ``dir`` (and parents), if necessary.

    Also supports ``name`` kwarg for specifying the exact basename of the temporary directory.
    """
    ctx0 = tmp_ensure_dir(dir) if dir else nullcontext()
    ctx1 = named_tmpdir(name, dir=dir, **kwargs)
    with (
        ctx0,
        ctx1 as rv,
    ):
        yield rv


tmpdir = TmpDir


@contextmanager
def TmpPath(name: str | None = None, **kwargs):
    """Yield a path to a temporary file with a given name."""
    name = name or 'tmpfile'
    with TemporaryDirectory(**kwargs) as tmpdir:
        path = join(tmpdir, name)
        parent = dirname(path)
        makedirs(parent, exist_ok=True)
        yield path


tmppath = TmpPath



from __future__ import annotations

import sys
from dataclasses import asdict, dataclass
from typing import Union

import pytest

from utz.dataclasses import from_dict
from utz.test import raises


@dataclass
class A1:
    s: str
    n: int
    b: bool | None = None


@dataclass
class B1:
    f: float
    arr: list[A1]
    obj: dict[str, A1]


@dataclass
class Obj1:
    obj: dict[int, A1 | None] | None = None


@dataclass
class Arr1:
    arr: list[A1 | None] | None = None


@dataclass
class A2:
    s: str
    n: int
    b: Union[bool, None] = None


@dataclass
class B2:
    f: float
    arr: list[A2]
    obj: dict[str, A2]


@dataclass
class Obj2:
    obj: Union[dict[int, Union[A2, None]], None] = None


@dataclass
class Arr2:
    arr: Union[list[Union[A2, None]], None] = None


def check(dc):
    assert from_dict(dc.__class__, asdict(dc)) == dc


@pytest.mark.parametrize(
    "A,B,Obj,Arr",
    [
        *([(A1, B1, Obj1, Arr1)] if sys.version_info >= (3, 10) else []),
        (A2, B2, Obj2, Arr2),
    ]
)
def test_roundtrips(A: type, B: type, Obj: type, Arr: type):
    check(A('aa', 11))
    check(A('aa', 11, True))
    check(
        B(
            1.11,
            [
                A('aa', 11),
                A('bb', 22),
            ],
            {
                'ccc': A('cc', 33),
                'ddd': A('dd', 44),
            }
        )
    )
    check(B(2.22, [], {}))
    check(Obj())
    check(Obj({111: A('aa', 11), 222: None}))
    check(Arr())
    check(Arr([A('aa', 11), None, A('bb', 22)]))


def test_bad_dicts():
    # Python 3.9 doesn't include the "A." at the start
    with raises(TypeError, r"(?:A.)?__init__\(\) missing 1 required positional argument: 'n'"):
        from_dict(A2, { 's': 'aa' })

    with raises(KeyError, "'c'"):
        from_dict(A2, { 's': 'aa', 'n': 11, 'c': 'invalid key' })

    with raises(ValueError, ["Invalid value '!!!' (expected one of ['bool', 'NoneType'])"]):
        from_dict(A2, { 's': 'aa', 'n': 11, 'b': '!!!' })

    with raises(ValueError, ["111 (int) is not of expected type float"]):
        from_dict(B2, { 'f': 111, 'arr': [], 'obj': {} })


class C:
    s: str
    n: int

    def __init__(self, s, n):
        self.s = s
        self.n = n


@dataclass
class D:
    c: C


def test_non_dataclasses():
    with raises(TypeError, '`fields` must be called with a dataclass type or instance, not C'):
        from_dict(C, { 's': 'aa', 'n': 11 })

    with raises(TypeError, '`fields` must be called with a dataclass type or instance, not C'):
        from_dict(D, { 'c': { 's': 'aa', 'n': 11 } })



from utz.case import dash_case, snake_case, camel_case

dash = 'aaa-bb-c'
snake = 'aaa_bb_c'
camel = 'AaaBbC'


def test_dash_case():
    assert dash_case(dash) == dash
    assert dash_case(snake) == dash
    assert dash_case(camel) == dash


def test_snake_case():
    assert snake_case(dash) == snake
    assert snake_case(snake) == snake
    assert snake_case(camel) == snake


def test_camel_case():
    assert camel_case(dash) == camel
    assert camel_case(snake) == camel
    assert camel_case(camel) == camel



from io import TextIOWrapper

import struct
from contextlib import contextmanager
from gzip import GzipFile


class DeterministicGzipFile(GzipFile):
    """
    A GzipFile subclass that writes deterministic output by:
    1. Setting fixed metadata values in the header
    2. Using a consistent modification time
    3. Not including the filename in the header
    """
    def _write_gzip_header(self, compresslevel):
        self.fileobj.write(b'\037\213')             # Magic number
        self.fileobj.write(b'\010')                 # Compression method
        fname = b''
        flags = 0
        self.fileobj.write(chr(flags).encode('latin-1'))
        mtime = 0  # Using 0 as a fixed modification time
        self.fileobj.write(struct.pack("<L", mtime))
        # Set extra flags based on compression level:
        # 2 -> maximum compression
        # 4 -> fastest algorithm
        xfl = 2 if compresslevel >= 9 else (4 if compresslevel == 1 else 0)
        self.fileobj.write(bytes([xfl]))
        self.fileobj.write(b'\377')                 # OS (unknown)
        if fname:
            self.fileobj.write(fname + b'\000')


@contextmanager
def deterministic_gzip_open(
    path: str,
    mode: str,
    compression_level: int = 9,
    encoding: str = 'utf-8',
):
    """
    Opens a gzip file with deterministic output settings.
    Note: The file object must be used within a context manager.
    """
    gzip_file = None
    raw_file = None
    try:
        raw_file = open(path, "wb")
        gzip_file = DeterministicGzipFile(
            fileobj=raw_file,
            mode=mode,
            compresslevel=compression_level,
        )
        # Wrap in TextIOWrapper if text mode is requested
        if 'b' not in mode:
            gzip_file = TextIOWrapper(gzip_file, encoding=encoding)
        yield gzip_file
    finally:
        if gzip_file:
            gzip_file.close()
        if raw_file:
            raw_file.close()



from dataclasses import dataclass
from os.path import dirname, join

from utz import hash_file, HashName, parametrize

TEST_DIR = dirname(__file__)
ROOT_DIR = dirname(TEST_DIR)


@dataclass
class case:
    hash_name: HashName
    expected_hash: str
    file: str = 'src/utz/hash.py'

    @property
    def id(self):
        return f"{self.hash_name}({self.file})"


@parametrize(
    case('sha256', '26f8c940220e1d3f81a99eb46b3856cc94704336c679cd2e0d3b61ee5f47ce41'),
    case('sha384', '8d55b5142a80dc7a95298158eeaa4adf72054d681f7b51016380def5307e359b6719163678c463c55abc071b82e7da33'),
    case('sha3_256', '1310c83274fb6194eb781f5c6b49a6db0cba8968ea8f92f08c738290186090b9'),
    case('sha3_512', 'cef055919d9bc6ccf7c2e641a9b4f1ccfc3efca9409d1b96a3cf4be70257f0acc96b7e3c5619763793a87a901ff4fe8f3676e2125e4502f7aa88307e8d760e0e'),
    case('md5', '75fb9145b5f5dc52cc26be44489808c6'),
)
def test_hash_file(hash_name: HashName, expected_hash: str, file: str):
    path = join(ROOT_DIR, file)
    actual_hash = hash_file(path, hash_name=hash_name)
    assert actual_hash == expected_hash



from __future__ import annotations

from asyncio import gather, set_event_loop
from os import getcwd, remove, makedirs
from os.path import splitext, exists, dirname

import asyncio
import json
import memray
from contextlib import AbstractContextManager, AbstractAsyncContextManager
from subprocess import DEVNULL
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import Literal, Self

from utz import err
from utz.aio import proc
from utz.process.log import Log, silent


Verbosity = Literal[0, 1, 2]


class Tracker(AbstractContextManager, AbstractAsyncContextManager):
    def __init__(
        self,
        path: str | None = None,
        keep: bool | None = None,
        native_traces: bool = True,
        follow_fork: bool = True,
        log: Log | bool | Verbosity = None,
        stats: bool = True,
        flamegraph: bool = True,
        **kwargs,
    ):
        if not stats and not flamegraph and not keep:
            raise ValueError("At least one of {stats, flamegraph, keep} should be True")
        self.path = path
        self.keep = keep
        self.verbose = 1
        if log is True:
            log = err
        elif isinstance(log, int) and log > 0:
            self.verbose = log
            log = err
        elif not log:
            log = silent
            self.verbose = 0
        self.log = log
        self.tmpfile = False
        self.kwargs = dict(
            native_traces=native_traces,
            follow_fork=follow_fork,
            **kwargs,
        )
        self.peak_mem = None
        self.stats = None
        self.compute_stats = stats
        self.compute_flamegraph = flamegraph
        self.tracker = None

    def __enter__(self) -> Self:
        self.peak_mem = self.stats = None
        assert not self.tracker, f"Attempted to `__enter__` MemTracker before `__exit__`ing"
        path = self.path
        if path is None:
            path = self.path = NamedTemporaryFile(dir=getcwd(), suffix=".memray").name
            self.tmpfile = True
        elif exists(path):
            self.log(f"mem.Tracker removing {path}")
            remove(path)

        makedirs(dirname(path), exist_ok=True)
        self.log(f"memray logging to {path}")
        self.tracker = memray.Tracker(path, **self.kwargs)
        self.tracker.__enter__()
        return self

    async def __aenter__(self) -> Self:
        return self.__enter__()

    @property
    def stats_path(self) -> str | None:
        path = self.path
        if path is None:
            return path
        else:
            return f'{splitext(self.path)[0]}.stats.json'

    @property
    def flamegraph_path(self) -> str | None:
        path = self.path
        if path is None:
            return path
        else:
            return f'{splitext(self.path)[0]}.html'

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        self.tracker.__exit__(exc_type, exc_value, exc_tb)
        self.tracker = None
        rm = self.keep is False or (self.keep is None and self.tmpfile)
        coros = []
        if self.compute_stats:
            stats_path = self.stats_path
            if not exc_value:
                coros.append(
                    proc.run(
                        'memray', 'stats', '--json', '-fo', stats_path, self.path,
                        log=self.log,
                        **(dict() if self.verbose > 1 else dict(stdout=DEVNULL)),
                    )
                )
        if self.compute_flamegraph:
            flamegraph_path = self.flamegraph_path
            if not exc_value:
                coros.append(
                    proc.run(
                        'memray', 'flamegraph', '-fo', flamegraph_path, self.path,
                        log=self.log,
                        **(dict() if self.verbose > 1 else dict(stdout=DEVNULL)),
                    )
                )

        await gather(*coros)
        if rm:
            remove(self.path)
            self.path = None
        if exc_value:
            raise exc_value
        with open(stats_path, 'r') as f:
            stats = json.load(f)
        if rm:
            remove(stats_path)
        self.stats = stats
        self.peak_mem = stats['metadata']['peak_memory']

        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        coro = self.__aexit__(exc_type, exc_value, exc_tb)
        try:
            # If we're already in an event loop
            loop = asyncio.get_running_loop()
            loop.run_until_complete(coro)
        except RuntimeError:
            # If no event loop is running, create one
            loop = asyncio.new_event_loop()
            set_event_loop(loop)
            try:
                loop.run_until_complete(coro)
            finally:
                loop.close()



import asyncio
from asyncio import sleep

from contextlib import contextmanager, asynccontextmanager

from pytest import raises
from tempfile import NamedTemporaryFile

from utz import exists, AbstractContextManager, contexts, acontexts, Time


class EnterException(Exception): pass
class ExitException(Exception): pass


class TestCtx(AbstractContextManager):
    instances = []

    def __init__(self, n, raise_on=None):
        self.n = n
        self.state = 'init'
        self.raise_on = raise_on
        self.instances += [self]

    def __enter__(self):
        self.state = 'entering'
        if self.raise_on == 'enter':
            raise EnterException
        self.state = 'entered'
        return self.n

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.state = 'exiting'
        if self.raise_on == 'exit':
            raise ExitException
        self.state = 'exited'


def test_contexts_tempfiles():
    files = [
        NamedTemporaryFile(suffix='111'),
        NamedTemporaryFile(suffix='222'),
        NamedTemporaryFile(suffix='333'),
    ]
    with contexts(files) as fds:
        assert [ exists(f.name) for f in fds ] == [True] * len(files)
        assert [ f.name for f in files ] == [ f.name for f in fds ]

    assert [ exists(f.name) for f in files ] == [False] * len(files)


def test_contexts_complex():
    f1 = NamedTemporaryFile()
    f2 = NamedTemporaryFile()
    f3 = NamedTemporaryFile()
    for f in [ f1, f2, f3 ]:
        assert exists(f.name)

    with raises(EnterException):
        with contexts(
            f1,
            TestCtx(111),
            TestCtx(222),
            TestCtx(333, raise_on='enter'),
            TestCtx(444),
            f2,
            TestCtx(555, raise_on='exit'),
            TestCtx(666),
            f3,
        ) as (F1, c1, c2, c3, c4, F2, c5, c6, F3):
            raise AssertionError  # we never get here bc c3 raises during __enter__

    # NamedTemporaryFile() creates temp file during __init__, not __enter__! Unfortunate…
    assert [ exists(f.name) for f in [ f1, f2, f3 ] ] == [ False, True, True, ]
    f2.close()
    f3.close()
    assert [ exists(f.name) for f in [ f1, f2, f3 ] ] == [ False, False, False, ]

    assert { i.n: i.state for i in TestCtx.instances } == {
        111: 'exited',
        222: 'exited',
        333: 'entering',
        444: 'init',
        555: 'init',
        666: 'init',
    }


n = 2


@contextmanager
def double():
    global n
    n = n * 2
    yield


@contextmanager
def to_str():
    global n
    n = str(n)
    yield


def test_contexts_order():
    global n
    n = 2
    with contexts(double(), to_str()) as (_n, _s):
        assert n == '4'
    n = 2
    with contexts(to_str(), double()) as (_s, _n):
        assert n == '22'


@asynccontextmanager
async def exit_sleep(s: float):
    try:
        yield
    finally:
        await sleep(s)


async def _test_acontexts(time: Time):
    async with acontexts(
        exit_sleep(0.1),
        exit_sleep(0.1),
        exit_sleep(0.1),
    ):
        with time("inside"):
            pass


def test_acontexts():
    time = Time()
    with time("acontexts"):
        asyncio.run(_test_acontexts(time))

    inside = time.times["inside"]
    assert 0 < inside < 1e-5
    total = time.times["acontexts"]
    assert 0.1 < total < 0.11



from __future__ import annotations

import pytest
import typing
from click import command, argument
from click.testing import CliRunner
from functools import partial
from typing import Literal

from utz.cli import cmd, count, num, obj, flag, parse_int, multi, inc_exc, incs, multi_cb, excs
from utz.rgx import Patterns, Includes

parametrize = pytest.mark.parametrize

Verbosity = Literal["warn", "info", "debug"]
Verbosities = typing.get_args(Verbosity)


def _check(
    cli,
    args: list[str],
    *expected_lines: str,
    exit_code: int = 0,
    msg: str | None = None,
):
    runner = CliRunner()
    res = runner.invoke(cli, args)
    actual_lines = res.output.split('\n')
    if not actual_lines[-1]:
        actual_lines.pop()

    expected = [ list(expected_lines), exit_code ]
    actual = [ actual_lines, res.exit_code ]
    if exit_code == 0 and res.exit_code != 0:
        raise res.exception
    if msg is not None:
        expected.append((msg,))
        actual.append(res.exception.args)
    assert actual == expected


def checks(cli):
    return partial(_check, cli), partial(_check, cli, exit_code=2)


def test_count_enum():
    @command
    @count('-v', '--verbosity', values=Verbosities, help="0x: warn, 1x: info, 2x: debug")
    def count_cli(verbosity: Verbosity):
        print(verbosity)

    check, fail = checks(count_cli)
    check([], "warn")
    check(['-v'], 'info')
    check(['-vv'], 'debug')
    check(['-v', '-v'], 'debug')
    fail(
        ['-vvv'],
        "Usage: count-cli [OPTIONS]",
        "Try 'count-cli --help' for help.",
        "",
        "Error: Invalid value for '-v' / '--verbosity': expected [0,3), found 3",
    )


def test_count():
    @command
    @count('-v', '--verbosity', help="0x: warn, 1x: info, 2x: debug")
    def count_cli(verbosity: int):
        print(verbosity)

    check, _ = checks(count_cli)
    check([], '0')
    check(['-v'], '1')
    check(['-vv'], '2')
    check(['-v', '-v'], '2')
    check(['-vvv'], '3')


def test_num_opt():
    @command
    @num('-n', '--num')
    def num_opt_cli(num: int | None):
        if num is not None:
            assert isinstance(num, int)
        print(num)

    check, fail = checks(num_opt_cli)
    check([], 'None')
    check(['-n', '123'], '123')
    check(['-n10k'], '10000')
    check(['-n10Ki'], '10240')
    check(['-n1.0123m'], '1012300')
    check(['-n1Gi'], str(2**30))
    fail(
        ['-n1bi'],
        "Usage: num-opt-cli [OPTIONS]",
        "Try 'num-opt-cli --help' for help.",
        "",
        "Error: Invalid value for '-n' / '--num': failed to parse \"1bi\"",
    )


def test_num():
    @command
    @num('-n', '--num', default=111)
    def num_cli(num: int):
        print(num)

    check, fail = checks(num_cli)
    check(['-n', '123'], '123')
    check(['-n10k'], '10000')
    check(['-n10Ki'], '10240')
    check(['-n1.0123m'], '1012300')
    check([], '111')


def test_num_required():
    @command
    @num('-n', '--num', required=True)
    def num_cli(num: int):
        print(num)

    check, fail = checks(num_cli)
    check(['-n', '123'], '123')
    check(['-n10k'], '10000')
    check(['-n10Ki'], '10240')
    check(['-n1.0123m'], '1012300')
    fail(
        [],
        'Usage: num-cli [OPTIONS]',
        "Try 'num-cli --help' for help.",
        '',
        "Error: Missing option '-n' / '--num'.",
    )


def test_obj():
    @cmd
    @obj('-e', '--env')
    @flag('-v', '--verbose')
    def obj_cli(env: dict[str, str], verbose: bool):
        for k, v in env.items():
            if verbose:
                print(f'{k=} {v=}')
            else:
                print(f'{k}={v}')

    check, fail = checks(obj_cli)
    check(['-e', 'NUM=111'], 'NUM=111')
    check(['-e', 'NUM=111', '-eSTR=abc'], 'NUM=111', 'STR=abc')
    check([])
    check(['--env', '==', '-e===', '-v'], "k='' v='=='")
    fail(
        ['-e', '123'],
        "Usage: obj-cli [OPTIONS]",
        "Try 'obj-cli --help' for help.",
        "",
        "Error: Invalid value for '-e' / '--env': bad value: 123",
    )


def test_obj_parse():
    @command
    @obj('-n', '--num', 'nums', parse=parse_int)
    @flag('-v', '--verbose')
    def obj_cli(nums: dict[str, int], verbose: bool):
        for k, v in nums.items():
            if verbose:
                print(f'{k=} {v=}')
            else:
                print(f'{k}={v:,}')

    check, fail = checks(obj_cli)
    check(['-n', 'lo=10k'], 'lo=10,000')
    check(['-n', 'lo=111', '-nhi=222'], 'lo=111', 'hi=222')
    check([])
    fail(['--num', '='], exit_code=1, msg='Failed to parse ""')
    fail(
        ['-n', '123'],
        "Usage: obj-cli [OPTIONS]",
        "Try 'obj-cli --help' for help.",
        "",
        "Error: Invalid value for '-n' / '--num': bad value: 123",
    )


def test_multi():
    @command
    @multi('-n', '--num', 'nums', parse=parse_int)
    def obj_cli(nums: tuple[int, ...]):
        assert isinstance(nums, tuple)
        if nums:
            print(f"{', '.join(f'{n:,}' for n in nums)}")

    check, fail = checks(obj_cli)
    check(['-n', '10k,1m,111'], '10,000, 1,000,000, 111')
    check(['-n', '10k,1m', '--num', '111'], '10,000, 1,000,000, 111')
    # check(['-n', 'lo=111', '-nhi=222'], 'lo=111', 'hi=222')
    check([])
    fail(
        ['--num', ''],
        "Usage: obj-cli [OPTIONS]",
        "Try 'obj-cli --help' for help.",
        "",
        "Error: Invalid value for '-n' / '--num': failed to parse \"\"",
    )
    fail(
        ['-n', '111', '--num', '12a'],
        "Usage: obj-cli [OPTIONS]",
        "Try 'obj-cli --help' for help.",
        "",
        "Error: Invalid value for '-n' / '--num': failed to parse \"12a\"",
    )


# Example strings to test vs. regexs /a./ and /b/ below
matches = ['aa', 'bc', 'cb']
misses = ['c', 'a', 'AA', 'B']
both = matches + misses


def test_inc_exc():
    @command
    @inc_exc(
        multi('-i', '--include', help="Print arguments iff they match at least one of these regexs; comma-delimited, and can be passed multiple times"),
        multi('-x', '--exclude', help="Print arguments iff they don't match any of these regexs; comma-delimited, and can be passed multiple times"),
    )
    @argument('vals', nargs=-1)
    def cli(patterns: Patterns, vals: tuple[str, ...]):
        for val in vals:
            assert patterns.pats != []  # patterns.pats should be `None` when no patterns are passed, which also implies `not bool(patterns)`
            if patterns(val):
                print(val)

    check, fail = checks(cli)
    check(both, *both)
    check(['-i', 'a.,b', *both], *matches)
    check(['-i', 'a.', '--include', 'b', *both], *matches)
    check(['-x', 'a.,b', *both], *misses)
    check(['-x', 'a.', '--exclude', 'b', *both], *misses)
    fail(['-i', 'a', '-x', 'b'], msg='Pass -i/--include xor -x/--exclude', exit_code=1)


@parametrize(
    'deco,args,expected', [
        *[
            (deco, args, expected)
            for deco in [
                # `incs` can receive a fully-constructed `click` decorator, e.g. using the `multi` helper:
                incs(multi('-i', '--include', 'patterns', help="Print arguments iff they match at least one of these regexs; comma-delimited, and can be passed multiple times")),
                # `incs` can also receive args/kwargs that will be passed to `click.option`, to construct a decorator; this example recreates what `multi` does above:
                incs('-i', '--include', 'patterns', multiple=True, callback=multi_cb, help="Print arguments iff they match at least one of these regexs; comma-delimited, and can be passed multiple times"),
            ]
            for args, expected in [
                ([*both], both),
                ([ '-i', 'a.,b', *both], matches),
                ([ '-i', 'a.', '--include', 'b', *both], matches),
                ([ '-i', '[A-Z]', 'AAA', 'bbb', 'CCC', 'ddd', ], ['AAA', 'CCC']),
            ]
        ],
        *[
            (
                # Simple, non-"multi" version:
                incs('-i', '--include', 'patterns', help="Print arguments iff they match at least one of these regexs; comma-delimited"),
                args, expected,
            ) for args, expected in [
                ([ '-i', '[A-Z]', 'AAA', 'bbb', 'CCC', 'ddd', ], ['AAA', 'CCC']),
                ([ 'AAA', 'bbb', 'CCC', 'ddd', ], [ 'AAA', 'bbb', 'CCC', 'ddd', ]),
            ]
        ],
        *[
            (deco, args, expected)
            for deco in [
                # Similarly, `excs` can receive a fully-constructed `click` decorator, e.g. using the `multi` helper:
                excs(multi('-x', '--exclude', 'patterns', help="Print arguments iff they don't match any of these regexs; comma-delimited, and can be passed multiple times")),
                # `excs` can also receive args/kwargs that will be passed to `click.option`, to construct a decorator; this example recreates what `multi` does above:
                excs('-x', '--exclude', 'patterns', multiple=True, callback=multi_cb, help="Print arguments iff they don't match any of these regexs; comma-delimited, and can be passed multiple times"),
            ]
            for args, expected in [
                ([*both], both),
                ([ '-x', 'a.,b', *both], misses),
                ([ '-x', 'a.', '--exclude', 'b', *both], misses),
                ([ '-x', '[A-Z]', 'AAA', 'bbb', 'CCC', 'ddd', ], ['bbb', 'ddd']),
            ]
        ],
        *[
            (
                # Simple, non-"multi" version:
                excs('-x', '--exclude', 'patterns', help="Print arguments iff they don't match any of these regexs; comma-delimited"),
                args, expected,
            ) for args, expected in [
                ([ '-x', '[A-Z]', 'AAA', 'bbb', 'CCC', 'ddd', ], ['bbb', 'ddd']),
                ([ 'AAA', 'bbb', 'CCC', 'ddd', ], [ 'AAA', 'bbb', 'CCC', 'ddd', ]),
            ]
        ],
    ],
)
def test_incs(deco, args, expected):
    @command
    @deco
    @argument('vals', nargs=-1)
    def cli(patterns: Patterns, vals: tuple[str, ...]):
        for val in vals:
            if patterns(val):
                print(val)

    check, _ = checks(cli)
    check(args, *expected)


def test_incs_arg():
    @command
    @argument('val_str')
    # `incs` can even accept a `click.argument`; in this case, multiple comma-delimited strings are accepted:
    @incs(argument('includes', callback=multi_cb, nargs=-1))
    def cli(val_str: str, includes: Includes):
        for val in val_str.split(' '):
            if includes(val):
                print(val)

    check, _ = checks(cli)
    check([' '.join(both), 'a.', 'b'], *matches)
    check([' '.join(both), 'a.,b'], *matches)



from os import getcwd, makedirs
from os.path import basename, join, exists, realpath, isdir, dirname
from tempfile import TemporaryDirectory

from utz import tmp_ensure_dir, TmpDir, TmpPath

cwd = getcwd()


def test_TmpDir():
    with TmpDir(dir='.') as subdir:
        name = basename(subdir)
        assert realpath(subdir) == join(cwd, name)
        assert exists(subdir)
    assert not exists(subdir)

    with TmpDir(dir='abc') as subdir:
        name = basename(subdir)
        assert realpath(subdir) == join(cwd, 'abc', name)
        assert exists(subdir)
    assert not exists(subdir)

    with TmpDir('name', dir='abc') as subdir:
        assert basename(subdir) == 'name'
        slug = basename(dirname(subdir))
        assert realpath(subdir) == join(cwd, 'abc', slug, 'name')
        assert exists(subdir)
    assert not exists(subdir)

    with TmpDir('name') as subdir:
        assert basename(subdir) == 'name'
        parent = dirname(subdir)
        assert exists(subdir)
    assert not exists(subdir)
    assert not exists(parent)


def test_tmp_ensure_dir():
    with TemporaryDirectory(dir='.') as tmpdir:
        name = basename(tmpdir)
        assert realpath(tmpdir) == join(cwd, name)
        dir1 = join(name, 'aaa')
        makedirs(dir1)
        dir2 = join(dir1, 'bbb')
        with tmp_ensure_dir(dir2):
            assert exists(dir2)
            assert isdir(dir2)
        assert not exists(dir2)
        assert exists(dir1)

        dir3 = join(dir2, 'ccc')
        with tmp_ensure_dir(dir3):
            assert exists(dir2)
            assert exists(dir3)
            assert isdir(dir2)
            assert isdir(dir3)
        assert not exists(dir2)
        assert not exists(dir3)
        assert exists(dir1)
    assert not exists(dir1)


def test_TmpPath():
    name = 'foo.txt'
    with TmpPath(name, dir='.') as path:
        assert basename(path) == name
        assert not exists(path)
        parent = dirname(path)
        slug = basename(parent)
        assert exists(parent)
        assert isdir(parent)
        assert realpath(path) == join(cwd, slug, name)
    assert not exists(parent)

    relpath = join('dir1', 'dir2', name)
    with TmpPath(relpath, dir='.') as path:
        assert basename(path) == name
        assert not exists(path)
        dir2 = dirname(path)
        assert basename(dir2) == 'dir2'
        dir1 = dirname(dir2)
        assert basename(dir1) == 'dir1'
        parent = dirname(dir1)
        slug = basename(parent)
        assert exists(parent)
        assert isdir(parent)
        assert realpath(path) == join(cwd, slug, relpath)
    assert not exists(parent)



from __future__ import annotations

from contextlib import nullcontext, contextmanager
from dataclasses import dataclass
from os import makedirs
from os.path import dirname, exists, relpath
from shutil import rmtree
from tempfile import NamedTemporaryFile
from typing import Any

from utz import parametrize


def fn(f: float, fmt: str) -> str:
    """Example function, to be tested with ``case``s below."""
    return f"{f:{fmt}}"


@dataclass
class case:
    """Container for a test-case; float, format, and expected output."""
    f: float
    fmt: str
    expected: str

    @property
    def id(self):
        return f"fmt-{self.f}-{self.fmt}"


@parametrize(
    case(1.23, "0.1f", "1.2"),
    case(123.456, "0.1e", "1.2e+02"),
    case(-123.456, ".0f", "-123"),
)
def test_fn(f, fmt, expected):
    """Example test, "parametrized" by several ``Cases``s."""
    assert fn(f, fmt) == expected


@parametrize(
    case(1.23, "0.1f", "1.2"),
)
def test_case_id(case, request):
    """Test that the ``id`` property of a ``Case`` is used as the test's ID.

    Also, an argument named ``case`` is populated with the full ``Case`` object.

    ``request`` is a special arg-name that Pytest populates with metadata about the current test case.
    """
    assert case.f == 1.23
    assert case.fmt == "0.1f"
    assert case.id == "fmt-1.23-0.1f"
    assert request.node.name == 'test_case_id[fmt-1.23-0.1f]'


def pow2(n: int) -> int:
    """Example function to be tested by ``case2``s below."""
    return n ** 2


@dataclass
class case2:
    n: int
    # Example field with a default value, omitted from test IDs unless set to a non-default value
    extra: str = "extra str"

    @property
    def sq(self):
        return self.n * self.n


@parametrize(
    case2(11),
    case2(22, extra="extra str"),
)
def test_case2_id(case, request):
    """Test that an ``id`` is autogenerated for "case" classes that don't define one.

    Fields are included iff their value doesn't match the default value (or there is no default value).
    """
    assert request.node.name == f'test_case2_id[{case.n}]'


@parametrize(
    case2(11, extra="AAAA"),
)
def test_case2_id_extra(request):
    """Test that autogenerated ``id``s incorporate non-default field values."""
    assert request.node.name == 'test_case2_id_extra[]'


@parametrize(
    case2(n)
    for n in range(10)
)
def test_pow2(n, sq):
    """Example unpacking a `@property` field (``sq``) from a "case" class."""
    assert pow2(n) == sq


def json_dump(obj: Any, path: str):
    """Example function to be tested by ``case3``s below."""
    import json
    with open(path, 'w') as f:
        json.dump(obj, f)


def json_load(path: str) -> Any:
    """Example function to be tested by ``case3``s below."""
    import json
    with open(path, 'r') as f:
        return json.load(f)


@dataclass
class case3:
    obj: Any  # Object to JSON-serde
    dir: str | None = None  # Perform JSON-serde roundtrip in this directory
    suffix: str | None = None  # Suffix for the temporary file

    @property
    @contextmanager
    def dir_ctx(self):
        if not self.dir:
            with nullcontext():
                yield
        else:
            rm_dir = False
            if not exists(self.dir):
                makedirs(self.dir)
                rm_dir = True
            try:
                yield self.dir
            finally:
                if rm_dir:
                    rmtree(self.dir)

    def roundtrip(self) -> Any:
        """JSON serialization round-trip."""
        with self.dir_ctx:
            with NamedTemporaryFile(dir=self.dir, suffix=self.suffix) as f:
                path = f.name
                if self.dir:
                    assert dirname(relpath(path)) == self.dir
                json_dump(self.obj, path)
                return json_load(path)


@parametrize(
    case3({'a': 1}),
    case3([ 11, 22, 33 ]),
    dir=[ None, 'tmp/test_json_roundtrip', ],
    suffix=[ None, ".json", ],
    # Custom field-formatters for pytest "ID" strings; see ``ids``
    id_fmts={
        'obj': lambda obj: (
            ','.join(f'{k}={v}' for k, v in obj.items())
            if isinstance(obj, dict)
            else str(obj).replace(' ', '')
        ),
        ('dir', 'suffix'): lambda s: str(s) if s else None,  # None ⟹ omit from ID
    },
)
def test_json_roundtrip_sweeps(obj, roundtrip, request):
    """Example testing a function that writes and reads a JSON file.

    Also reflects over the "swept" (kwargs) arguments above, verifying that 8 test cases are
    generated, having the expected IDs."""
    assert obj == roundtrip()

    # Swept test-case IDs
    ids = [
        item.name
        for item in request.session.items
        if item.originalname == 'test_json_roundtrip_sweeps'
    ]
    # Current test-case ID
    cur = request.node.name
    assert cur in ids
    assert ids == [
        'test_json_roundtrip_sweeps[a=1]',
        'test_json_roundtrip_sweeps[a=1-.json]',
        'test_json_roundtrip_sweeps[a=1-tmp/test_json_roundtrip]',
        'test_json_roundtrip_sweeps[a=1-tmp/test_json_roundtrip-.json]',
        'test_json_roundtrip_sweeps[[11,22,33]]',
        'test_json_roundtrip_sweeps[[11,22,33]-.json]',
        'test_json_roundtrip_sweeps[[11,22,33]-tmp/test_json_roundtrip]',
        'test_json_roundtrip_sweeps[[11,22,33]-tmp/test_json_roundtrip-.json]',
    ]


@parametrize(
    case3('aaa', dir='tmp/test_json_roundtrip'),
)
def test_contextmanager(case, dir_ctx):
    """Example testing a context manager."""
    assert not exists(case.dir)
    with dir_ctx as tmpdir:
        assert exists(case.dir)
        assert case.dir == tmpdir
    assert not exists(case.dir)



from __future__ import annotations

from contextlib import contextmanager
from os import chdir, getcwd
from pathlib import Path

from utz.tmpdir import TmpDir


class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, path: str | Path, mk: bool = False):
        path = Path(path)
        self.path = path.expanduser()
        self.mk = mk

    def __enter__(self):
        self.prevPath = Path(getcwd())
        if self.mk:
            self.path.mkdir(parents=True, exist_ok=True)
        chdir(str(self.path))

    def __exit__(self, etype, value, traceback):
        chdir(str(self.prevPath))


@contextmanager
def cd_tmpdir(*args, **kwargs):
    """Create a ``TemporaryDirectory`` (creating any parents of ``dir``, as necessary) and ``cd`` into it."""
    with TmpDir(*args, **kwargs) as rv, cd(rv):
        yield rv



import os
import shutil
import tempfile
from contextlib import contextmanager


@contextmanager
def named_pipes(n: int = 1) -> list[str]:
    """Yield a list of paths to named pipes that are created and destroyed

    From https://stackoverflow.com/a/28840955"""
    dirname = tempfile.mkdtemp()
    try:
        paths = [os.path.join(dirname, 'named_pipe' + str(i)) for i in range(n)]
        for path in paths:
            os.mkfifo(path)
        yield paths
    finally:
        shutil.rmtree(dirname)




def i2s(v, chars):
    N = len(chars)
    base = 1
    bases = []
    ceiling = 1
    while ceiling <= v:
        bases.append(base)
        base *= N
        ceiling += base

    chrs = ''
    n = v - ceiling
    for base in reversed(bases):
        i = n // base
        n -= i*base
        ch = chars[i]
        chrs += ch

    return chrs


def s2i(s, ints):
    sz = len(s)
    N = len(ints)
    n = (N**sz-1)//(N-1)
    base = 1
    for ch in reversed(s):
        i = ints[ch]
        n += i*base
        base *= N
    return n


def b2s(b, chars):
    n = 0
    for byte in b:
        n *= 256
        n += byte
    return i2s(n, chars)


class Converter:
    """Convert between non-negative integers and strings comprised of a given alphabet.

    Several instantiations follow, but note that e.g. ``b64`` differs from standard
    base64-encoding; ``Converter``s provide isomorphisms between all natural numbers and all
    strings (given an "alphabet"):
    - 0 ⟺ ''
    - [1,N) ⟺ <1-char strings>
    - [N, N²+N) ⟺ <2-char strings>
    - [N²+N, N³+N²+N) ⟺ <3-char strings>
    - etc.

    They also preserve ordering; alphabets are ASCII-ordered.
    """
    i2s: str

    def __init__(self, i2s: str):
        self.i2s = i2s
        self.s2i = { ch: i for i, ch in enumerate(self.i2s) }

    def __call__(self, v):
        if isinstance(v, str):
            return s2i(v, self.s2i)
        elif isinstance(v, int):
            return i2s(v, self.i2s)
        elif isinstance(v, bytes):
            return b2s(v, self.i2s)
        else:
            raise ValueError(f'Unrecognized type ({type(v)}): {v}')


b26u = Converter('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
b26l = Converter('abcdefghijklmnopqrstuvwxyz')
b36u = Converter('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ')
b36l = Converter('0123456789abcdefghijklmnopqrstuvwxyz')
b52 = Converter('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
b62 = Converter('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789')
b64 = Converter('+/0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
b90 = Converter('!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz')



from functools import wraps

from utz import recvs, call


def fn1(a, b):
    return dict(a=a, b=b)


@wraps(fn1)
def fn2(a, c, **kwargs):
    return dict(a=a, c=c, kwargs=kwargs)


def test_recvs():
    assert recvs(fn1, 'a')
    assert recvs(fn1, 'b')
    assert not recvs(fn1, 'c')

    assert recvs(fn2, 'a')
    assert recvs(fn2, 'b')
    assert recvs(fn2, 'c')
    assert not recvs(fn2, 'd')


def test_call():
    assert call(fn1, a=1, b=2) == dict(a=1, b=2)
    assert call(fn1, a=1, b=2, c=3) == dict(a=1, b=2)

    assert call(fn2, a=1, b=2, c=3) == dict(a=1, c=3, kwargs=dict(b=2))
    assert call(fn2, a=1, b=2, c=3, d=4) == dict(a=1, c=3, kwargs=dict(b=2))



"""
"Export" most members of this module, as well as the standard library (c/o `stdlb`). Add:

```python
from utz import *
```

to notebooks (or your `$PYTHONSTARTUP` script) to save lots of `import` boilerplate.
"""

# Optional-import helper
from .imports import _try

from os import environ
from .environ import env

# Import most of the Python standard library
# `_try` wrap helps ensure `utz[setup]` can install properly
with _try: from stdlb import *

# ### Date/Time
with _try: from dateutil.parser import parse
with _try: from pytz import UTC
from .time import now, today, Time, utc

# ### Jupyter
with _try: from IPython.display import HTML, Image, Markdown, display

# Import other utilities from this repo:

from .backoff import backoff
from .bases import b26u, b26l, b36u, b36l, b52, b62, b64, b90
from .hash import hash_file, HashName
from .path import mkdir, mkpar
from .rgx import Patterns, Includes, Excludes
from os import path
from .tmpdir import named_tmpdir, tmpdir, TmpDir, tmp_ensure_dir, tmppath, TmpPath

from . import proc, process
from .process import *

from .escape import esc

from . import fn
from .fn import args, call, decos, recvs

from .gzip import deterministic_gzip_open, DeterministicGzipFile

from . import jsn
from .jsn import Encoder

from . import size
from .size import iec

with _try:
    from . import pnds
    from .pnds import *

from .cd import cd, cd_tmpdir

from . import docker
from .docker.dsl import *

# `utz.Yield` is used by some modules below, e.g. `utz.ym`
from .context import *

from .o import o, rev
from .use import use
from .ym import YM
from .ymd import YMD

with _try:
    from git import Git, Repo
    from .git import make_repo

from .git import github

with _try:
    from .collections import coerce, only, is_subsequence, singleton, solo, one, e1

from .defaultdict import DefaultDict, Unset

from .test import parametrize, raises

# ## Optional Modules

# joblib: easy parallelization
with _try:
    from joblib import Parallel, delayed
from .parallel import parallel

try:
    from . import plots
    from .plots import plot
    import plotly.express as px
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except ImportError:
    pass

# yaml
with _try:
    import yaml
    # Fix a bad default in PyYAML (cf. https://github.com/yaml/pyyaml/issues/110)
    from functools import partial
    yaml.dump = partial(yaml.dump, sort_keys=False)
    yaml.safe_dump = partial(yaml.safe_dump, sort_keys=False)

# requests
with _try:
    from requests import \
        get as   GET, \
        post as  POST, \
        put as   PUT, \
        patch as PATCH

# ## PyData / Scientific Python

with _try:
    import numpy as np
    from numpy import concatenate, array, ndarray, matrix, nan

with _try: import seaborn as sns
with _try: import matplotlib.pyplot as plt
with _try: from scipy.sparse import spmatrix, coo_matrix, csr_matrix, csc_matrix

from .version import git_version, pkg_version



from contextlib import contextmanager
from os.path import basename
from re import match
from sys import stderr
from traceback import print_exc
from typing import Iterable

from utz import now, tmpdir
from utz.proc import check, line, run


@contextmanager
def tmp(
        url,
        *clone_args,
        branch=None,
        ref=None,
        submodules=True,
        push=False,
        pull=False,
        cd=True,
        name=None,
        bare=False,
        dir=None,
        **run_kwargs,
):
    """contextmanager for creating a Git repo in a temporary directory, and optionally cd'ing into it and upstreaming
    commits from it

    - `url`: local or remote path to Git repository to be cloned
    - `clone_args` (list[str]): passed directly to `git clone`
    - `branch`: Git branch to operate on inside temporary clone (by default, inherit from origin)
    - `ref`: if `branch` doesn't already exist, reset it to this ref. Note: if `ref` is omitted, `branch` must already
        exist in the remote! If `ref` is set and `branch` isn't, a temporary/nonce branch name will be used.
    - `submodules`: recursively clone submodules (`git clone --recurse-submodules`)
    - `push` (str | bool | list[str]): run a `git push` after `yield`ing; `str` or `list[str]` serve as arguments to
        `git push`
    - `pull` (bool): upstream changes post-`yield` by `cd`ing into the origin directory (must be a local dir, otherwise
        use `push`) and running a `git pull` from this temporary clone
    - `cd` (bool): move into the temporary clone dir before `yield`ing
    - `name`: basename for the temporary clone directory (defaults to basename of `url`)
    - `bare`: clone a bare repository
    """
    import utz
    from utz import git
    name = name or basename(url)
    if name.endswith('.git'):
        name = name[:-len('.git')]
    with tmpdir(name, dir=dir) as repo_dir:
        cmd = ['git', 'clone']
        if submodules:
            cmd += ['--recurse-submodules']
        if bare:
            cmd += ['--bare']
        if branch and not ref:
            # branch must already exist and we want to clone and work on it
            cmd += ['-b', branch]
        cmd += clone_args
        cmd += [url, repo_dir, ]
        run(*cmd, **run_kwargs)
        if ref:
            with utz.cd(repo_dir):
                if ref is True:
                    ref = git.branch.current()

                if not branch:
                    branch = f'tmp-{now("short")}'

                if git.branch.exists(branch):
                    make_branch = False
                else:
                    make_branch = True

                if make_branch:
                    if bare:
                        run('git', 'branch', branch, ref)
                        run('git', 'symbolic-ref', 'HEAD', f'refs/heads/{branch}')
                    else:
                        run('git', 'checkout', '-b', branch, ref)

                if cd:
                    yield repo_dir
        if cd:
            if not ref:
                # if `checkout or init`, we've already cd'd and yielded above
                with utz.cd(repo_dir):
                    yield repo_dir
        else:
            yield repo_dir

        # optionally cd into upstream `url` dir (assumed to be a local directory if `pull` is set) and pull changes
        # from the temporary `repo_dir`
        if pull:
            with utz.cd(url):
                if not branch:
                    branch = git.branch.current()
                run('git', 'fetch', repo_dir)
                try:
                    ref_line = line('git', 'ls-remote', repo_dir, f'refs/heads/{branch}')
                    m = match(
                        r'(?P<sha>[0-9a-f]{40})\s+refs/heads/%s' % branch,
                        ref_line,
                    )
                    if m:
                        sha = m['sha']
                    else:
                        raise RuntimeError(f'Unrecognized ls-remote line: {ref_line}')
                except RuntimeError:
                    stderr.write('%s\n' % f'Error fetching branch {branch} from remote {repo_dir}')
                    print_exc()
                    sha = git.sha('FETCH_HEAD')

                if not check('git', 'merge', '-m', '\n'.join([f'Merge {branch} from tmpdir', '', repo_dir]), sha):
                    tmp_branch = f'conflict-{sha}'
                    run('git', 'branch', tmp_branch, sha)
                    stderr.write('Pulling from %s failed, aborting merge, saving to %s\n' % (repo_dir, tmp_branch))
                    run('git', 'merge', '--abort')

        # optionally push back upstream
        if push:
            with utz.cd(repo_dir):
                cmd = [ 'git', 'push', ]
                remote = line('git', 'remote')
                if isinstance(push, str):
                    cmd += [ remote, f'{push}:{push}', ]
                elif isinstance(push, Iterable):
                    cmd += push
                elif push is True:
                    if branch:
                        cmd += [ remote, f'{branch}:{branch}', ]
                run(*cmd)



import pytest

from test.paths import GSMO, HAILSTONE
from utz import basename, CalledProcessError, cd, env, exists, getcwd, git, lines, realpath, run, line, \
    now, b62


def test_current_branch():
    branch = 'tmp'
    sha = 'e0add3d'
    with git.clone.tmp(HAILSTONE, branch=branch, ref=sha):
        assert git.branch.current() == branch
        assert git.sha() == sha
        run('git', 'checkout', sha)  # detach HEAD
        with pytest.raises(CalledProcessError):
            assert git.branch.current() == branch  # by default, no branch ⟹ error
        assert git.branch.current(sha_ok=True) == sha  # can return SHA instead
        assert git.branch.current(sha_ok=None) is None  # or None


def check(cwd=None, name=None, status=True, branch=None, paths=None, shas=None, files=None, rm=None, ):
    if cwd:
        # Check that we're actually in the directory returned by the ctx mgr above:
        dir = realpath(getcwd())
        assert dir == realpath(cwd)

    if name:
        # check basename
        assert basename(getcwd()) == name

    if status:
        if status is True:
            # verify clone is clean
            assert not lines('git', 'status', '--short')
        else:
            # verify status lines
            assert lines('git', 'status', '--short') == status

    # verify current branch
    if branch:
        assert git.branch.current() == branch

    # verify submodules have been cloned
    if paths:
        if isinstance(paths, str): paths = [paths]
        for path in paths:
            assert exists(path), f"{path} doesn't exist"

    # verify various shas
    for sha, refs in shas.items():
        if isinstance(refs, (list, tuple)):
            for ref in refs:
                if ref is None:
                    assert git.sha() == sha
                else:
                    assert git.sha(ref) == sha

    if files:
        for path, lns in files.items():
            assert lines('cat', path) == lns

    if rm:
        assert not exists(rm)


def commit_file(path, lines, parent_sha=None, mode='w', fmt='%h', commit=True):
    new_file = not exists(path)
    with open(path, mode) as f:
        if isinstance(lines, str): lines = [lines]
        f.writelines('%s\n' % ln for ln in lines)
    run('git', 'add', path)
    if new_file:
        msg = f'add {path}: {lines}'
    else:
        msg = f'update {path}: {lines}'
    if commit:
        run('git', 'commit', '-m', msg)
        if parent_sha:
            assert git.sha('HEAD^') == parent_sha
        return git.fmt(fmt)


def test_tmp_clone_remote_push_changes():
    # set this HAILSTONE_SSH_URL to a different fork or repo if developing without access to this one
    url = env.get('HAILSTONE_SSH_URL', 'git@gitlab.com:gsmo/examples/hailstone.git')
    nonce = b62(now().ms)
    branch = f'tmp-{nonce}'
    sha0 = 'e0add3d'
    with git.clone.tmp(url, branch=branch, ref=sha0, push=True) as cwd:
        tmpdir = cwd
        check(cwd=cwd, name='hailstone', shas={sha0: (None, branch)})
        sha1 = commit_file('test1.txt', ['111', '222'], sha0, fmt='%H')

    try:
        assert not exists(tmpdir)
        assert git.ls_remote(url, head=branch) == sha1
    finally:
        run('git', 'push', '--delete', url, branch)


def test_tmp_clone_local_pull_changes():
    branch = 'tmp'
    tag = 'v0.1.1rc2'
    sha0 = 'de6da9d'

    # bind some `check` params
    def verify(wd, sha):
        check(
            cwd=wd,
            name='gsmo',
            paths='example/hailstone/run.ipynb',
            branch=branch,
            shas={sha: (None, branch)},
        )

    with git.clone.tmp(GSMO, branch=branch, ref=tag) as origin:
        verify(origin, sha0)

        # simulate an additional clone + pulling changes back in
        with git.clone.tmp(origin, branch=branch, pull=True) as wd:
            tmpdir = wd
            verify(wd, sha0)
            sha1 = commit_file('test1.txt', ['123', '456'], sha0)

        check(
            shas={sha1: (None, branch)},
            branch=branch,
            files={'test1.txt': ['123', '456']},
            rm=tmpdir,  # verify the tmpdir is gone
        )

        # simulate an additional clone + merging changes back in with concurrent upstream changes
        with git.clone.tmp(origin, branch=branch, pull=True) as wd:
            tmpdir = wd
            verify(wd, sha1)
            sha2_2 = commit_file('test2.txt', ['abc', 'def'], sha1)

            # meanwhile, make another change in origin (which will have to be merged w/ sha2)
            with cd(origin):
                sha2_1 = commit_file('test1.txt', '789', sha1, mode='a')

            # on ctx mgr exit, merge is performed…

        sha2 = git.sha()
        check(
            shas={
                sha2: branch,
                sha2_1: 'HEAD^',
                sha2_2: 'HEAD^2',
            },
            branch=branch,
            files={
                'test1.txt': ['123', '456', '789', ],
                'test2.txt': ['abc', 'def', ],
            },
            rm=tmpdir,
        )

        # simulate an additional clone + failing to merge changes in vs concurrent upstream changes, but leaving a ref
        # for safe-keeping
        with git.clone.tmp(origin, branch=branch, pull=True) as wd:
            tmpdir = wd
            verify(wd, sha2)
            sha3_2 = commit_file('test2.txt', 'zzz', sha2, mode='a')

            # meanwhile, make another change in origin (which will have to be merged w/ sha2)
            with cd(origin):
                sha3_1 = commit_file('test2.txt', 'yyy', sha2, mode='a')

            # on ctx mgr exit, merge is performed, but a merge conflict arises, and a new branch `conflict-<sha>` points
            # at the unmerged temporary/remote changes

        conflict_branch = f'conflict-{sha3_2}'
        check(
            shas={
                sha3_1: (None, branch),
                sha3_2: conflict_branch,
                sha2: 'HEAD^',
            },
            branch=branch,
            files={
                'test1.txt': ['123', '456', '789', ],
                'test2.txt': ['abc', 'def', 'yyy', ],
            },
            rm=tmpdir,
        )

    # verify the tmpdir is gone
    assert not exists(origin)


def test_bare_tmp_clone():
    branch = 'tmp'
    sha0 = 'e0add3d'
    with git.clone.tmp(HAILSTONE, bare=True, branch=branch, ref=sha0) as origin:
        check(cwd=origin, status=False, branch=branch, shas={sha0: (None, branch)})
        with git.clone.tmp(origin, branch=branch) as wd:
            check(cwd=wd, name='hailstone', shas={sha0: (None, branch)})
            sha1 = commit_file('test1.txt', ['aaa', 'bbb'], sha0)
            remote = line('git', 'remote')
            run('git', 'push', remote)
            check(
                branch=branch,
                shas={
                    sha1: (None, branch, f'{remote}/{branch}'),
                    sha0: 'HEAD^',
                },
                files={'test1.txt': ['aaa', 'bbb']},
            )
        check(
            branch=branch,
            status=False,  # bare clone!
            shas={
                sha1: (None, branch),
                sha0: 'HEAD^',
            },
        )


def test_ls_remote_lines():
    from utz.git.remote import parse_ls_remote_lines
    sha = 'e0add3d2805fc8999dab650697a22f1939fd5396'
    hailstone_lines = [
        f'{sha}	HEAD',
        f'{sha}	refs/heads/master',
    ]
    assert parse_ls_remote_lines(hailstone_lines) == dict(head=sha, heads={'master': sha})
    assert parse_ls_remote_lines(hailstone_lines, sha=sha) == dict(head=sha, heads={'master': sha})
    assert parse_ls_remote_lines(hailstone_lines, head='master') == sha

    utz_lines = [
        '12bbfa076261df4ed8069bb91044971bd47892a8	HEAD',
        '3aded57969e3e71d2f28c47ed328529fb84fe963	refs/heads/master',
        'aec6bf6cb37f02d9e0c0fe9f136973f2adc221e3	refs/heads/nb',
        '12bbfa076261df4ed8069bb91044971bd47892a8	refs/heads/py',
        'a5dfc290398236c11ecc6afe576e70027793eb1a	refs/tags/v0.1.3',
        '5313cee06bc686c90bb5dd6a2e2c3c8ad2d27a4a	refs/tags/v0.2.0',
        'a67870037d4c3006d515244ce9e9c51a6345e175	refs/tags/v0.2.1',
        '12bbfa076261df4ed8069bb91044971bd47892a8	refs/tags/v0.2.2',
    ]
    head = '12bbfa076261df4ed8069bb91044971bd47892a8'
    heads = {
        'master': '3aded57969e3e71d2f28c47ed328529fb84fe963',
        'nb': 'aec6bf6cb37f02d9e0c0fe9f136973f2adc221e3',
        'py': '12bbfa076261df4ed8069bb91044971bd47892a8',
    }
    tags = {
        'v0.1.3': 'a5dfc290398236c11ecc6afe576e70027793eb1a',
        'v0.2.0': '5313cee06bc686c90bb5dd6a2e2c3c8ad2d27a4a',
        'v0.2.1': 'a67870037d4c3006d515244ce9e9c51a6345e175',
        'v0.2.2': '12bbfa076261df4ed8069bb91044971bd47892a8',
    }
    assert parse_ls_remote_lines(utz_lines) == dict(head=head, heads=heads, tags=tags, )
    assert parse_ls_remote_lines(utz_lines, sha=head) == dict(head=head, heads={'py': head}, tags={'v0.2.2': head}, )
    assert parse_ls_remote_lines(utz_lines, head='py') == head
    assert parse_ls_remote_lines(utz_lines, head='nope') is None
    assert parse_ls_remote_lines(utz_lines, tag='v0.2.2') == head
    assert parse_ls_remote_lines(utz_lines, tag='nope') is None
    assert parse_ls_remote_lines(utz_lines, heads=True) == heads
    assert parse_ls_remote_lines(utz_lines, tags=True) == tags
    assert parse_ls_remote_lines(utz_lines, heads=True, tags=True) == dict(heads=heads, tags=tags)
    assert parse_ls_remote_lines(utz_lines, sha=head, heads=True) == {'py': head}
    assert parse_ls_remote_lines(utz_lines, sha=head, tags=True) == {'v0.2.2': head}
    assert parse_ls_remote_lines(utz_lines, sha=head, heads=True, tags=True) == dict(heads={'py': head},
                                                                                     tags={'v0.2.2': head})

    assert git.ls_remote('https://gitlab.com/runsascoded/utz.git', tag='v0.2.2') == head


def test_atom():
    branch = 'tmp'
    sha0 = 'e0add3d'
    with git.clone.tmp(HAILSTONE, branch=branch, ref=sha0):
        with git.txn():
            pass
        check(shas={
            sha0: (None, branch),
        })

        with git.txn(add='value') as txn:
            commit_file('value', ['6'], commit=False)
            msg = 'set value=6'
            txn.msg = msg
        sha1 = git.sha()
        assert git.msg() == msg
        check(
            shas={
                sha1: (None, branch),
                sha0: f'{branch}^',
            },
            files={'value': ['6']},
        )

        # no additional commit is created with `add` param configured
        with git.txn():
            sha2 = commit_file('value', ['3'])
        sha3 = git.sha()
        assert git.msg() == f'merge txn: {sha1}, {sha2}'
        check(
            shas={
                sha3: (None, branch),
                sha2: f'{branch}^2',
                sha1: f'{branch}^',
            },
            files={'value': ['3']}
        )

        msg = 'two more updates'
        with git.txn(msg=msg):
            sha4 = commit_file('value', ['10'])
            sha5 = commit_file('value', ['5'])
        sha6 = git.sha()
        assert git.msg() == msg
        check(shas={
            sha6: (None, branch),
            sha5: f'{branch}^2',
            sha4: f'{branch}^2^',
            sha2: f'{branch}^',
        })

        msg = 'set value=8'
        with git.txn(add='value') as txn:
            sha7 = commit_file('value', ['16'])
            commit_file('value', ['8'], commit=False)
            txn.msg = msg
        sha8 = git.sha()
        assert git.msg() == msg
        check(
            shas={
                sha8: (None, branch),
                sha7: f'{branch}^2',
                sha6: (f'{branch}^2^', f'{branch}^',),
            },
            files={'value': ['8']},
        )



from __future__ import annotations

import re

from typing import Sequence

from re import Pattern

from abc import ABC


class Patterns(ABC):
    def __init__(
        self,
        pats: Sequence[str | Pattern] | None,
        flags: int = 0,
        *,
        search: bool = True,
    ):
        self.pats = None if pats is None else [
            re.compile(pat, flags=flags) if isinstance(pat, str) else
            re.compile(pat.pattern, flags=flags) if flags else
            pat
            for pat in pats
        ]
        self.search = search

    def __bool__(self) -> bool:
        return self.pats is not None

    def __call__(self, val: str) -> bool:
        pats = self.pats
        if not self:
            # `pats is None` ⟹ everything passes this filter
            return True
        else:
            return any(pat.search(val) if self.search else pat.fullmatch(val) for pat in pats)


class Includes(Patterns):
    pass


class Excludes(Patterns):
    def __call__(self, val: str) -> bool:
        if not self:
            return True
        else:
            return not super().__call__(val)



# S3 utilities:
#
# - `client()`: cached boto3 S3 client
# - `parse_bkt_key(args: tuple[str, ...]) -> tuple[str, str]`: parse bucket and key from s3:// URL or separate arguments
# - `get_etag(*args: str, err_ok: bool = False, strip: bool = True) -> str | None`: get ETag of S3 object
# - `get_etags(*args: str) -> dict[str, str]`: get ETags for all objects with the given prefix
# - `atomic_edit(...) -> Iterator[str]`: context manager for atomically editing S3 objects

from __future__ import annotations

import re
from contextlib import contextmanager
from functools import cache
from os import getcwd, path
from os.path import exists, join
from tempfile import TemporaryDirectory
from typing import Iterator

import boto3
from botocore.exceptions import ClientError

from utz import Log, err, silent


@cache
def client():
    return boto3.client('s3')


def parse_bkt_key(args: tuple[str, ...]) -> tuple[str, str]:
    if len(args) == 1:
        arg = args[0]
        if arg.startswith('s3://'):
            arg = arg[len('s3://'):]
        bkt, key = arg.split('/', 1)
    elif len(args) == 2:
        bkt, key = args
    else:
        raise ValueError('Too many arguments')
    return bkt, key


def get_etag(
    *args: str,
    err_ok: bool = False,
    strip: bool = True,
) -> str | None:
    """
    Get the ETag of an S3 object.

    Args:
        args (str): The full s3:// URL of the object, or the bucket and key as separate arguments
        err_ok (bool): If True, return None instead of raising FileNotFoundError if object doesn't exist
        strip (bool): If True, strip quotes from ETag

    Returns:
        str: The ETag value of the S3 object, or None if it doesn't exist (and err_ok=True)
    """
    bkt, key = parse_bkt_key(args)
    s3 = client()

    try:
        res = s3.head_object(Bucket=bkt, Key=key)
        etag = res.get('ETag', '')
        if strip:
            etag = etag.strip('"')
        return etag
    except ClientError as e:
        if e.response['Error']['Code'] != '404':
            raise
        if not err_ok:
            raise FileNotFoundError(f"Object {bkt}/{key} does not exist")
        return None


def get_etags(*args: str) -> dict[str, str]:
    """Return etags for all objects with the given prefix.

    Args:
        args (str): The full s3:// URL of the object, or the bucket and key as separate arguments

    Returns:
        dict[str, str]: A mapping of object keys to ETags
    """
    bkt, key = parse_bkt_key(args)
    s3 = client()
    res = s3.list_objects_v2(Bucket=bkt, Prefix=key)
    etags = {
        obj['Key']: obj['ETag'].strip('"')
        for obj in res.get('Contents', [])
    }
    return etags


class ETagConflictError(RuntimeError):
    """ETag has changed, indicating conflicting updates"""
    pass


class ConditionalRequestError(RuntimeError):
    """Concurrent conflicting operation detected"""
    pass


@contextmanager
def atomic_edit(
    bucket_or_url: str,
    key: str | None = None,
    *,
    s3 = None,
    create_ok: bool = False,
    download: bool | None = None,
    rm_ok: bool = False,
    basename: str | None = None,
    keep: bool = False,
    log: Log = err,
    dry_run: bool = False,
    **kwargs
) -> Iterator[str]:
    """
    Context manager for atomically editing S3 objects with optimistic locking.
    Yields a path in a temporary directory. If a file is created at that path,
    it will be conditionally uploaded to S3 on exit.

    Args:
        bucket_or_url: Either bucket name or full s3:// URL
        key: Optional object key (required if bucket_or_url is bucket name)
        s3: Optional S3 client
        create_ok: If True, allows creation of new objects
        download: If True, download object to temp path before yielding
        rm_ok: If True, delete object if temp path is removed. Implies download=True, unless download=False is passed explicitly.
        basename: Optional name for temp file (defaults to key basename)
        keep: If True, don't delete temp file on exit
        log: Optional logger function
        dry_run: If True, don't actually upload (or delete) the modified file
        **kwargs: Additional arguments to pass to put_object

    Yields:
        Path in temporary directory where file can be created

    Raises:
        ETagConflictError: If ETag has changed (HTTP 412)
        ConditionalRequestError: If concurrent conflict (HTTP 409)
        ClientError: For other S3 errors
        FileNotFoundError: If create_ok=False and object doesn't exist
        ValueError: If URL is invalid or key is missing
    """
    log = log or silent

    if not s3:
        s3 = client()

    # Parse bucket and key from args
    if key is None:
        match = re.match(r's3://([^/]+)/(.+)', bucket_or_url)
        if not match:
            raise ValueError("Single argument must be s3:// URL")
        bkt, key = match.groups()
    else:
        bkt = bucket_or_url

    url = f"s3://{bkt}/{key}"
    # Get current etag if object exists
    etag0 = get_etag(bkt, key, err_ok=create_ok, strip=False)

    with TemporaryDirectory(dir=getcwd()) as tmpdir:
        tmp_path = join(tmpdir, basename or path.basename(key))
        if download or (rm_ok and download is not False):
            log(f"Downloading {url} to {tmp_path}")
            s3.download_file(bkt, key, tmp_path)

        yield tmp_path

        if exists(tmp_path):
            if etag0:
                kwargs['IfMatch'] = etag0

            try:
                if dry_run:
                    log(f"Dry run: would upload {tmp_path} to {url}")
                    etag1 = get_etag(bkt, key, err_ok=create_ok, strip=False)
                    if etag0 != etag1:
                        raise ETagConflictError(f"ETag mismatch: {etag0} != {etag1}")
                else:
                    with open(tmp_path, 'rb') as f:
                        s3.put_object(
                            Bucket=bkt,
                            Key=key,
                            Body=f,
                            **kwargs
                        )
            except ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == 'PreconditionFailed':
                    raise ETagConflictError("ETag mismatch - object was modified") from e
                if error_code == 'ConditionalRequestFailed':
                    raise ConditionalRequestError("Concurrent conflicting operation") from e
                raise
        else:
            if rm_ok:
                if dry_run:
                    log(f"Would delete {url}")
                else:
                    log(f"Deleting {url}")
                    s3.delete_object(Bucket=bkt, Key=key)



from __future__ import annotations

from json import loads

import asyncio
from subprocess import CalledProcessError, PIPE, CompletedProcess, DEVNULL

from utz.process import Cmd, err, Arg, Log, Json


async def run(
    *args: Arg,
    dry_run: bool = False,
    log: Log = err,
    check: bool = True,
    **kwargs,
) -> CompletedProcess | None:
    """Async convenience wrapper for subprocess execution."""
    cmd = Cmd.mk(*args, **kwargs)
    if dry_run:
        if log:
            log(f'Would run: {cmd}')
    else:
        args, kwargs = cmd.compile(log=log)
        proc = await asyncio.create_subprocess_exec(*args, **kwargs)
        await proc.wait()

        if check:
            if proc.returncode != 0:
                raise CalledProcessError(proc.returncode, args)
            return None
        else:
            # Creating CompletedProcess to match subprocess.run's return type
            return CompletedProcess(
                args=args,
                returncode=proc.returncode,
                # Only include stdout/stderr if they were captured
                stdout=None if proc.stdout is None else await proc.stdout.read(),
                stderr=None if proc.stderr is None else await proc.stderr.read(),
            )


sh = run


async def output(
    *args: Arg,
    dry_run: bool = False,
    both: bool = False,
    err_ok: bool | None = False,
    log: Log = err,
    **kwargs,
) -> bytes | None:
    """Async convenience wrapper for subprocess output capture.

    By default, logs commands to `err` (stderr) before running (pass `log=None` to disable).

    If ``err_ok=True``, exceptions will be caught and suppressed, and the stdout up to that point will be returned.
    If ``err_ok=None``, exceptions will be caught and suppressed, and ``None`` will be returned.

    ``both=True`` is an alias for ``stderr=STDOUT``.
    """
    cmd = Cmd.mk(*args, **kwargs)
    if dry_run:
        if log:
            log(f'Would run: {cmd}')
        return None
    else:
        args, kwargs = cmd.compile(log=log, both=both)
        try:
            if kwargs['shell']:
                # For shell=True, we need to join args into a single string
                # and use create_subprocess_shell
                proc = await asyncio.create_subprocess_shell(
                    args,
                    stdout=PIPE,
                    **kwargs
                )
            else:
                proc = await asyncio.create_subprocess_exec(
                    *args,
                    stdout=PIPE,
                    **kwargs
                )

            # Stream and capture the output in real-time
            output = b''
            assert proc.stdout  # for type checking
            while True:
                line = await proc.stdout.readline()
                if not line:
                    break
                output += line

            await proc.wait()
            if proc.returncode != 0:
                raise CalledProcessError(proc.returncode, cmd, output=output)
            else:
                return output
        except CalledProcessError as e:
            if err_ok is True:
                return e.output
            elif err_ok is None:
                return None
            else:
                raise e


async def text(*args, **kwargs) -> str | None:
    return (await output(*args, **kwargs)).decode()


async def json(
    *cmd: Arg,
    dry_run: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> Json:
    """Run a command, parse the output as JSON, and return the parsed object."""
    out = await output(*cmd, dry_run=dry_run, err_ok=err_ok, **kwargs)
    if out is None or err_ok is True and not out:
        return None
    return loads(out.decode())


async def check(
    *cmd: Arg,
    stdout=DEVNULL,
    stderr=DEVNULL,
    **kwargs,
):
    """Run a command, return True iff it runs successfully (i.e. exits with code 0)."""
    try:
        await run(*cmd, stdout=stdout, stderr=stderr, **kwargs)
        return True
    except CalledProcessError:
        return False


async def lines(
    *cmd: Arg,
    keep_trailing_newline: bool = False,
    dry_run: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> list[str] | None:
    """Return the lines written to stdout by a command."""
    out = await output(*cmd, dry_run=dry_run, err_ok=err_ok, **kwargs)
    if err_ok is None and out is None:
        return None

    lines = [
        line.rstrip('\n')
        for line in
        out.decode().split('\n')
    ]

    if not keep_trailing_newline and lines and not lines[-1]:
        lines = lines[:-1]

    return lines


async def line(
    *cmd: Arg,
    empty_ok: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> str | None:
    """Run a command, verify that it returns a single line of output, and return that line."""
    _lines = await lines(*cmd, err_ok=err_ok, **kwargs)
    if (empty_ok or err_ok is not False) and not _lines:
        return None
    elif len(_lines) == 1:
        return _lines[0]
    else:
        raise ValueError(f'Expected 1 line, found {len(_lines)}:\n\t%s' % '\n\t'.join(_lines))


__all__ = [
    'text',
    'json',
    'check',
    'line',
    'lines',
    'run',
    'output',
    'sh',
]
